from __future__ import annotations

from exo import Memory, DRAM, instr


def _is_const_size(sz, c):
    return sz.isdecimal() and int(sz) == c


def _is_some_const_size(sz):
    return sz.isdecimal() and int(sz) > 0


# --------------------------------------------------------------------------- #
#   Neon registers
# --------------------------------------------------------------------------- #


class Neon(Memory):
    @classmethod
    def global_(cls):
        return "#include <arm_neon.h>"

    @classmethod
    def can_read(cls):
        return False

    @classmethod
    def alloc(cls, new_name, prim_type, shape, srcinfo):
        if not shape:
            raise MemGenError(f"{srcinfo}: Neon vectors are not scalar values")

        vec_types = {
            "float": (4, "float32x4_t"),
            "double": (2, "float64x2_t"),
            "_Float16": (8, "float16x8_t"),
        }

        if not prim_type in vec_types.keys():
            raise MemGenError(f"{srcinfo}: Neon vectors must be f32/f64 (for now)")

        reg_width, C_reg_type_name = vec_types[prim_type]

        if not _is_const_size(shape[-1], reg_width):
            raise MemGenError(
                f"{srcinfo}: Neon vectors of type {prim_type} must be {reg_width}-wide, got {shape}"
            )
        shape = shape[:-1]
        if shape:
            if not all(_is_some_const_size(s) for s in shape):
                raise MemGenError(
                    f"{srcinfo}: Cannot allocate variable numbers of Neon vectors"
                )
            result = f'{C_reg_type_name} {new_name}[{"][".join(map(str, shape))}];'
        else:
            result = f"{C_reg_type_name} {new_name};"

        return result

    @classmethod
    def free(cls, new_name, prim_type, shape, srcinfo):
        return ""

    @classmethod
    def window(cls, basetyp, baseptr, indices, strides, srcinfo):
        assert strides[-1] == "1"
        idxs = indices[:-1] or ""
        if idxs:
            idxs = "[" + "][".join(idxs) + "]"
        return f"{baseptr}{idxs}"


# --------------------------------------------------------------------------- #
#   f32 Neon intrinsics
# --------------------------------------------------------------------------- #

#
# Load, Store, Broadcast, FMAdd, Mul, Add?
#
# float32


@instr("*{result} += vaddvq_f32({x_data});")
def neon_assoc_reduce_add_instr_4xf32(result: f32 @ DRAM, x: [f32][4] @ Neon):
    for i in seq(0, 4):
        result += x[i]


@instr("{dst_data} = vld1q_f32(&{src_data});")
def neon_vld_4xf32(dst: [f32][4] @ Neon, src: [f32][4] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("vst1q_f32(&{dst_data}, {src_data});")
def neon_vst_4xf32(dst: [f32][4] @ DRAM, src: [f32][4] @ Neon):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("{dst_data} = vld1q_dup_f32(&{src_data});")
def neon_broadcast_4xf32(dst: [f32][4] @ Neon, src: [f32][1] @ DRAM):
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[0]


@instr("{dst_data} = vld1q_dup_f32({src_data});")
def neon_broadcast_4xf32_scalar(dst: [f32][4] @ Neon, src: f32 @ DRAM):
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = src


@instr("{dst_data} = vmovq_n_f32(0.0f);")
def neon_zero_4xf32(dst: [f32][4] @ Neon):
    assert stride(dst, 0) == 1

    for i in seq(0, 4):
        dst[i] = 0.0


@instr("{dst_data} = vaddq_f32({lhs_data}, {rhs_data});")
def neon_vadd_4xf32(dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 4):
        dst[i] = lhs[i] + rhs[i]


@instr("{dst_data} = vaddq_f32({src_data}, {dst_data});")
def neon_reduce_vadd_4xf32(dst: [f32][4] @ Neon, src: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] += src[i]


@instr("{dst_data} = vmulq_f32({lhs_data}, {rhs_data});")
def neon_vmul_4xf32(dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 4):
        dst[i] = lhs[i] * rhs[i]


@instr("{dst_data} = vmulq_f32({dst_data}, {rhs_data});")
def neon_vmul2_4xf32(dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 4):
        dst[i] = lhs[i] * rhs[i]


@instr("{dst_data} = vfmaq_laneq_f32({dst_data}, {lhs_data}, {rhs_data}, {lane});")
def neon_vfmla_4xf32_4xf32(
    dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][4] @ Neon, lane: index
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert lane >= 0
    assert lane < 4
    for i in seq(0, 4):
        dst[i] += lhs[i] * rhs[lane]


# This function uses an extra buffer for a beta=0 approach
@instr("{dst_data} = vfmaq_laneq_f32({b_data}, {lhs_data}, {rhs_data}, {lane});")
def neon_vfmla2_4xf32_4xf32(
    dst: [f32][4] @ Neon,
    b: [f32][4] @ Neon,
    lhs: [f32][4] @ Neon,
    rhs: [f32][4] @ Neon,
    lane: index,
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert stride(b, 0) == 1
    assert lane >= 0
    assert lane < 4
    for i in seq(0, 4):
        dst[i] = b[i] + lhs[i] * rhs[lane]


@instr("{dst_data} = vmlaq_f32({dst_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_4xf32_4xf32(
    dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][4] @ Neon
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 4):
        dst[i] += lhs[i] * rhs[i]


@instr("{dst_data} = vmlaq_f32({res_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_ex_4xf32_4xf32(
    dst: [f32][4] @ Neon,
    res: [f32][4] @ Neon,
    lhs: [f32][4] @ Neon,
    rhs: [f32][4] @ Neon,
):
    assert stride(dst, 0) == 1
    assert stride(res, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 4):
        dst[i] = res[i] + lhs[i] * rhs[i]


@instr("{dst_data} = vmlaq_n_f32({dst_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_4xf32_1xf32(
    dst: [f32][4] @ Neon, lhs: [f32][4] @ Neon, rhs: [f32][1] @ DRAM
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1

    for i in seq(0, 4):
        dst[i] += lhs[i] * rhs[0]


@instr("{dst_data} = vmlaq_n_f32({dst_data}, {rhs_data}, {lhs_data});")
def neon_vfmadd_1xf32_4xf32(
    dst: [f32][4] @ Neon, lhs: [f32][1] @ DRAM, rhs: [f32][4] @ Neon
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1

    for i in seq(0, 4):
        dst[i] += lhs[0] * rhs[i]


# -----------------------------------------------
# Load, Store, Broadcast, FMAdd, Mul, Add?
#
# float16


@instr("{dst_data} = vld1q_f16((float16_t *)&{src_data});")
def neon_vld_8xf16(dst: [f16][8] @ Neon, src: [f16][8] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[i]


@instr("vst1q_f16((float16_t *)&{dst_data}, {src_data});")
def neon_vst_8xf16(dst: [f16][8] @ DRAM, src: [f16][8] @ Neon):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[i]


@instr("{dst_data} = vld1q_dup_f16((float16_t *)&{src_data});")
def neon_broadcast_8xf16(dst: [f16][8] @ Neon, src: [f16][1] @ DRAM):
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = src[0]


@instr("{dst_data} = vmovq_n_f16(0.0f);")
def neon_zero_8xf16(dst: [f16][8] @ Neon):
    assert stride(dst, 0) == 1

    for i in seq(0, 8):
        dst[i] = 0.0


@instr("{dst_data} = vaddq_f16({lhs_data}, {rhs_data});")
def neon_vadd_8xf16(dst: [f16][8] @ Neon, lhs: [f16][8] @ Neon, rhs: [f16][8] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 8):
        dst[i] = lhs[i] + rhs[i]


@instr("{dst_data} = vmulq_f16({lhs_data}, {rhs_data});")
def neon_vmul_8xf16(dst: [f16][8] @ Neon, lhs: [f16][8] @ Neon, rhs: [f16][8] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 8):
        dst[i] = lhs[i] * rhs[i]


@instr("{dst_data} = vfmaq_laneq_f16({dst_data}, {lhs_data}, {rhs_data}, {lane});")
def neon_vfmla_8xf16_8xf16(
    dst: [f16][8] @ Neon, lhs: [f16][8] @ Neon, rhs: [f16][8] @ Neon, lane: index
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1
    assert lane >= 0
    assert lane < 8
    for i in seq(0, 8):
        dst[i] += lhs[i] * rhs[lane]


@instr("{dst_data} = vfmaq_f16({dst_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_8xf16_8xf16(
    dst: [f16][8] @ Neon, lhs: [f16][8] @ Neon, rhs: [f16][8] @ Neon
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 8):
        dst[i] += lhs[i] * rhs[i]


@instr("{dst_data} = vfmaq_f16({res_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_ex_8xf16_8xf16(
    dst: [f16][8] @ Neon,
    res: [f16][8] @ Neon,
    lhs: [f16][8] @ Neon,
    rhs: [f16][8] @ Neon,
):
    assert stride(dst, 0) == 1
    assert stride(res, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 8):
        dst[i] = res[i] + lhs[i] * rhs[i]


@instr("{dst_data} = vfmaq_n_f16({dst_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_8xf16_1xf16(
    dst: [f16][8] @ Neon, lhs: [f16][8] @ Neon, rhs: [f16][1] @ DRAM
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1

    for i in seq(0, 8):
        dst[i] += lhs[i] * rhs[0]


@instr("{dst_data} = vfmaq_n_f16({dst_data}, {rhs_data}, {lhs_data});")
def neon_vfmadd_1xf16_8xf16(
    dst: [f16][8] @ Neon, lhs: [f16][1] @ DRAM, rhs: [f16][8] @ Neon
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1

    for i in seq(0, 8):
        dst[i] += lhs[0] * rhs[i]


# TODO: Hack for procedure aliasing issue, can be deleted once we have
#      better way of handling aliasing
@instr("{dst_data} = {src_data};")
def neon_reg_copy_4xf32(dst: [f32][4] @ Neon, src: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = src[i]


@instr("{dst_data} = vnegq_f32({src_data});")
def neon_vneg_4xf32(dst: [f32][4] @ Neon, src: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 4):
        dst[i] = -src[i]


# --------------------------------------------------------------------------- #
#   f64 Neon intrinsics
# --------------------------------------------------------------------------- #


@instr("{dst_data} = vld1q_f64(&{src_data});")
def neon_vld_2xf64(dst: [f64][2] @ Neon, src: [f64][2] @ DRAM):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[i]


@instr("vst1q_f64(&{dst_data}, {src_data});")
def neon_vst_2xf64(dst: [f64][2] @ DRAM, src: [f64][2] @ Neon):
    assert stride(src, 0) == 1
    assert stride(dst, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[i]


@instr("{dst_data} = vld1q_dup_f64(&{src_data});")
def neon_broadcast_2xf64(dst: [f64][2] @ Neon, src: [f64][1] @ DRAM):
    assert stride(dst, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[0]


@instr("{dst_data} = vld1q_dup_f64({src_data});")
def neon_broadcast_2xf64_scalar(dst: [f64][2] @ Neon, src: f64 @ DRAM):
    assert stride(dst, 0) == 1

    for i in seq(0, 2):
        dst[i] = src


@instr("{dst_data} = vmlaq_f64({dst_data}, {lhs_data}, {rhs_data});")
def neon_vfmadd_2xf64_2xf64(
    dst: [f64][2] @ Neon, lhs: [f64][2] @ Neon, rhs: [f64][2] @ Neon
):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 2):
        dst[i] += lhs[i] * rhs[i]


@instr("{dst_data} = vmovq_n_f64(0.0f);")
def neon_zero_2xf64(dst: [f64][2] @ Neon):
    assert stride(dst, 0) == 1

    for i in seq(0, 2):
        dst[i] = 0.0


@instr("*{result} += vaddvq_f64({x_data});")
def neon_assoc_reduce_add_instr_2xf64(result: f64 @ DRAM, x: [f64][2] @ Neon):
    for i in seq(0, 2):
        result += x[i]


@instr("{dst_data} = vmulq_f64({lhs_data}, {rhs_data});")
def neon_vmul_2xf64(dst: [f64][2] @ Neon, lhs: [f64][2] @ Neon, rhs: [f64][2] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 2):
        dst[i] = lhs[i] * rhs[i]


@instr("{dst_data} = vaddq_f64({lhs_data}, {rhs_data});")
def neon_vadd_2xf64(dst: [f64][2] @ Neon, lhs: [f64][2] @ Neon, rhs: [f64][2] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(lhs, 0) == 1
    assert stride(rhs, 0) == 1

    for i in seq(0, 2):
        dst[i] = lhs[i] + rhs[i]


@instr("{dst_data} = vaddq_f64({src_data}, {dst_data});")
def neon_reduce_vadd_2xf64(dst: [f64][2] @ Neon, src: [f64][2] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 2):
        dst[i] += src[i]


# TODO: Also a hack
@instr("{dst_data} = {src_data};")
def neon_reg_copy_2xf64(dst: [f64][2] @ Neon, src: [f64][2] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[i]


@instr("{dst_data} = vnegq_f64({src_data});")
def neon_vneg_2xf64(dst: [f64][2] @ Neon, src: [f64][2] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 2):
        dst[i] = -src[i]


# --------------------------------------------------------------------------- #
#   f32 to f64 conversion
# --------------------------------------------------------------------------- #


@instr("{dst_data} = vcvt_f64_f32(vget_low_f32({src_data}));")
def neon_convert_f32_lower_to_f64(dst: [f64][2] @ Neon, src: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[i]


@instr("{dst_data} = vcvt_f64_f32(vget_high_f32({src_data}));")
def neon_convert_f32_upper_to_f64(dst: [f64][2] @ Neon, src: [f32][4] @ Neon):
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, 2):
        dst[i] = src[2 + i]



from __future__ import annotations  # make Python behave

from exo import proc
from exo.stdlib.scheduling import *

old_split = repeat(divide_loop)

# I'm going to define a 1D version of a standard convolutional layer (cf. CuDNN)
# K - number of output channels
# C - number of input channels
# W - length of the input signal/tensor
# R - width of the filter kernel
def gen_conv1d():
    @proc
    def conv1d(
        K: size,
        C: size,
        W: size,
        R: size,
        w: R[K, C, R],
        x: R[C, W],
        res: R[K, W],
    ):
        # zero out the result memory
        for k_init in seq(0, K):
            for i_init in seq(0, W):
                res[k_init, i_init] = 0.0

        # do the convolution
        for k in seq(0, K):
            for c in seq(0, C):
                for i in seq(0, W):
                    for r in seq(0, R):
                        if 0 <= i - r:
                            res[k, i] += w[k, c, r] * x[c, i - r]

    return conv1d


def test_im2col(golden):
    conv1d = gen_conv1d()

    # Let's start applying scheduling
    im2col_conv = rename(conv1d, "im2col_conv")
    im2col_conv = reorder_loops(im2col_conv, "i r")
    im2col_conv = bind_expr(im2col_conv, "x[c, i-r]", "y")

    # next, we can start to lift that allocation
    # up and out of the loop
    im2col_conv = autolift_alloc(im2col_conv, "y:R", 5, keep_dims=True)

    # Then, we can fission the loop correspondingly,
    # separating what is now a data-marshalling statement from
    # the actual compute statement in two subsequent
    # loop nests via fissioning
    im2col_conv = autofission(im2col_conv, im2col_conv.find("y[c,r,i] = _").after(), 5)

    # Now, in order to expose these two parts of the computation as
    # re-usable sub-procedures, we want a way to factor them out.
    im2col_conv, im2col = extract_subproc(im2col_conv, "for c in _: _", "im2col")
    im2col_conv, matmul = extract_subproc(im2col_conv, "for k in _: _", "matmul")

    # Given this factoring, we can then proceed
    # to schedule these sub-procedures themselves.
    tiled_matmul = rename(matmul, "tiled_matmul")
    # split the loops we want to tile together
    tiled_matmul = old_split(tiled_matmul, "k", 8, ["khi", "klo"], tail="cut")
    tiled_matmul = reorder_loops(tiled_matmul, "klo c #0")
    tiled_matmul = reorder_loops(tiled_matmul, "klo r #0")
    tiled_matmul = reorder_loops(tiled_matmul, "klo i #0")
    tiled_matmul = old_split(tiled_matmul, "c #0", 8, ["chi", "clo"], tail="cut")
    tiled_matmul = reorder_loops(tiled_matmul, "clo r #0")
    tiled_matmul = reorder_loops(tiled_matmul, "clo i #0")
    tiled_matmul = reorder_loops(tiled_matmul, "clo klo #0")
    tiled_matmul = old_split(tiled_matmul, "i #0", 8, ["ihi", "ilo"], tail="cut")
    tiled_matmul = reorder_loops(tiled_matmul, "ilo klo #0")
    tiled_matmul = reorder_loops(tiled_matmul, "ilo clo #0")

    # We can invoke another scheduling directive
    # to change which version of the matmul gets scheduled
    im2col_conv = call_eqv(im2col_conv, "matmul(_,_,_,_,_,_,_)", tiled_matmul)
    assert f"{im2col}\n{matmul}\n{im2col_conv}" == golden


"""
    filename = "test_im2col"

    f_pretty = open(os.path.join(TMP_DIR, filename + "_pretty.atl"), "w")
    f_pretty.write(str(im2col_conv))
    f_pretty.close()

    im2col.compile_c(TMP_DIR, filename)
"""



from __future__ import annotations

from exo import *
from exo.libs.memories import *
from exo.platforms.x86 import *
from exo.platforms.neon import *
from exo.syntax import *
from exo.API_cursors import *
from exo.stdlib.analysis import *


def get_children(proc, cursor=InvalidCursor(), lr=True):

    if isinstance(cursor, InvalidCursor):
        cursor = proc
    elif isinstance(cursor, (StmtCursor, ExprCursor)):
        cursor = proc.forward(cursor)

    def expr_children(expr):
        if isinstance(expr, (ReadCursor, WindowExprCursor)):
            yield from expr.idx()
        elif isinstance(expr, UnaryMinusCursor):
            yield expr.arg()
        elif isinstance(expr, BinaryOpCursor):
            yield expr.lhs()
            yield expr.rhs()
        elif isinstance(expr, BuiltInFunctionCursor):
            yield from expr.args()
        elif isinstance(expr, (LiteralCursor, ReadConfigCursor)):
            pass
        else:
            raise SchedulingError(
                f"Got an instance of {type(expr)} which is unsupported."
            )

    def stmt_children(stmt):
        if isinstance(stmt, AllocCursor):
            if stmt.is_tensor():
                yield from stmt.shape()
        elif isinstance(stmt, (AssignCursor, ReduceCursor)):
            yield from stmt.idx()
            yield stmt.rhs()
        elif isinstance(stmt, ForCursor):
            yield stmt.lo()
            yield stmt.hi()
            yield from stmt.body()
        elif isinstance(stmt, IfCursor):
            yield stmt.cond()
            yield from stmt.body()
            if not isinstance(stmt.orelse(), InvalidCursor):
                yield from stmt.orelse()
        elif isinstance(stmt, CallCursor):
            yield from stmt.args()
        elif isinstance(stmt, WindowStmtCursor):
            yield from stmt.idx()
        elif isinstance(stmt, AssignConfigCursor):
            yield stmt.rhs()
        elif isinstance(stmt, PassCursor):
            pass
        else:
            raise SchedulingError(
                f"Got an instance of {type(stmt)} which is unsupported."
            )

    def generator():
        if isinstance(cursor, (ExprListCursor, BlockCursor, tuple, list)):
            for c in cursor:
                yield c
        elif isinstance(cursor, Procedure):
            yield from cursor.args()
            yield cursor.body()
        elif isinstance(cursor, ArgCursor):
            if cursor.is_tensor():
                yield cursor.shape()
        elif isinstance(cursor, ExprCursor):
            yield from expr_children(cursor)
        elif isinstance(cursor, StmtCursor):
            yield from stmt_children(cursor)
        else:
            raise SchedulingError(
                f"Got an instance of {type(cursor)} which is unsupported."
            )

    if lr:
        yield from generator()
    else:
        children = list(generator())
        yield from children[::-1]


def _get_cursors(proc, cursor=InvalidCursor(), node_first=False, lr=True):

    if not isinstance(cursor, (InvalidCursor, ExprListCursor, BlockCursor)):
        cursor = proc.forward(cursor)

    def dfs(cursor):
        if node_first:
            yield cursor

        for child in get_children(proc, cursor, lr):
            yield from dfs(child)

        if not node_first:
            yield cursor

    return dfs(cursor)


def lrn(proc, cursor=InvalidCursor()):
    yield from _get_cursors(proc, cursor=cursor, node_first=False)


def nlr(proc, cursor=InvalidCursor()):
    yield from _get_cursors(proc, cursor=cursor, node_first=True)


def rln(proc, cursor=InvalidCursor()):
    yield from _get_cursors(proc, cursor=cursor, node_first=False, lr=False)


def nrl(proc, cursor=InvalidCursor()):
    yield from _get_cursors(proc, cursor=cursor, node_first=True, lr=False)


def lrn_stmts(proc, block=InvalidCursor()):
    yield from filter(lambda s: isinstance(s, StmtCursor), lrn(proc, cursor=block))


def nlr_stmts(proc, block=InvalidCursor()):
    yield from filter(lambda s: isinstance(s, StmtCursor), nlr(proc, cursor=block))


def get_symbols(proc, cursor=InvalidCursor()):
    for c in lrn(proc, cursor):
        if hasattr(c, "name"):
            yield c.name()


def get_declaration(proc, ctxt, name):
    ctxt = proc.forward(ctxt)

    if not isinstance(ctxt, StmtCursor):
        stmt = get_enclosing_stmt(proc, ctxt)
    else:
        stmt = ctxt

    for stmt in get_observed_stmts(stmt):
        if isinstance(stmt, AllocCursor) and stmt.name() == name:
            return stmt
    for arg in proc.args():
        if arg.name() == name:
            return arg
    raise SchedulingError("Declaration not found!")


def get_unique_names(proc):
    cnt = 0
    syms = set(get_symbols(proc))
    while cnt < 100:
        name = f"var{cnt}"
        cnt += 1
        if name in syms:
            continue
        yield name


def is_stmt(proc, stmt):
    stmt = proc.forward(stmt)
    return isinstance(stmt, StmtCursor)


def is_if(proc, if_c):
    if_c = proc.forward(if_c)
    return isinstance(if_c, IfCursor)


def is_loop(proc, loop):
    loop = proc.forward(loop)
    return isinstance(loop, ForCursor)


def check_is_loop(proc, loop):
    if not is_loop(proc, loop):
        raise TypeError(f"loop is not a {ForCursor}")


def is_loop_bounds_const(proc, loop):
    check_is_loop(proc, loop)
    loop = proc.forward(loop)
    return isinstance(loop.lo(), LiteralCursor) and isinstance(loop.hi(), LiteralCursor)


def loop_body_len(proc, loop):
    check_is_loop(proc, loop)
    loop = proc.forward(loop)
    return len(loop.body())


def is_single_stmt_loop(proc, loop):
    check_is_loop(proc, loop)
    loop = proc.forward(loop)
    return loop_body_len(proc, loop) == 1


# --------------------------------------------------------------------------- #
# High-level cursor navigation functions
# --------------------------------------------------------------------------- #


class CursorNavigationError(Exception):
    pass


def match_depth(cursor, cursor_to_match):
    """
    Lifts [cursor] through the AST until [cursor] and [cursor_to_match] are at
    the same level, e.g. have the same parent. Returns an InvalidCursorError if
    [cursor_to_match]'s parent is not an ancestor of [cursor].
    """
    assert (
        cursor.proc() == cursor_to_match.proc()
    ), "cursors originate from different procs"

    while cursor.parent() != cursor_to_match.parent():
        cursor = cursor.parent()
        if isinstance(cursor, InvalidCursor):
            raise CursorNavigationError(
                "cursor_to_match's parent is not an ancestor of cursor"
            )

    return cursor


def get_top_level_stmt(proc, c):
    c = proc.forward(c)

    while not isinstance(c.parent(), InvalidCursor):
        c = c.parent()
    return c


def get_enclosing_scope(proc, cursor, scope_type):
    cursor = proc.forward(cursor)
    cursor = cursor.parent()
    while not isinstance(cursor, (scope_type, InvalidCursor)):
        cursor = cursor.parent()

    if isinstance(cursor, InvalidCursor):
        raise SchedulingError("No enclosing scope found")

    return cursor


def get_enclosing_loop(proc, cursor, n=1):
    cursor = proc.forward(cursor)
    for i in range(n):
        cursor = get_enclosing_scope(proc, cursor, ForCursor)
    return cursor


def get_enclosing_loop_by_name(proc, cursor: Cursor, loop_iter: str):
    """
    Gets the enclosing loop with the given iteration variable [loop_iter].
    """
    cursor = proc.forward(cursor)

    while not (isinstance(cursor, ForCursor) and cursor.name() == loop_iter):
        cursor = cursor.parent()
        if isinstance(cursor, InvalidCursor):
            raise CursorNavigationError("no enclosing loop found")

    return cursor


def get_enclosing_if(proc, cursor, n=1):
    cursor = proc.forward(cursor)
    for i in range(n):
        cursor = get_enclosing_scope(proc, cursor, IfCursor)
    return cursor


def get_enclosing_stmt(proc, cursor, n=1):
    cursor = proc.forward(cursor)
    for i in range(n):
        cursor = get_enclosing_scope(proc, cursor, StmtCursor)
    return cursor


def get_my_stmt(proc, cursor):
    cursor = proc.forward(cursor)
    if isinstance(cursor, StmtCursor):
        return cursor
    return get_enclosing_stmt(proc, cursor)


def get_index_in_body(proc, stmt, from_top=True):
    stmt = proc.forward(stmt)
    index = 0
    while not isinstance(stmt.prev(), InvalidCursor):
        stmt = stmt.prev()
        index += 1
    if not from_top:
        parent = get_parent(proc, stmt)
        index = -(len(parent.body()) - index)
    return index


def get_parent(proc, stmt):
    stmt = proc.forward(stmt)
    parent = stmt.parent()
    if isinstance(parent, InvalidCursor):
        parent = proc
    return parent


def get_parents(proc, stmt, up_to=None):
    """
    Returns all ancestors of `stmt` (exclusive) up to `up_to` (inclusive).
    If `up_to` is `None`, returns ancestors up to the AST root.
    """
    stmt = proc.forward(stmt)

    if up_to is not None:
        up_to = proc.forward(up_to)

        assert up_to.is_ancestor_of(stmt)
        if up_to == stmt:
            return

        stmt = stmt.parent()
        while stmt != up_to:
            yield stmt
            stmt = stmt.parent()
        yield up_to
    else:
        stmt = stmt.parent()
        while not isinstance(stmt, InvalidCursor):
            yield stmt
            stmt = stmt.parent()


def get_nth_inner_loop(proc, loop, n):
    loop = proc.forward(loop)
    inner_loops = list(filter(lambda s: is_loop(proc, s), loop.body()))
    if n >= len(inner_loops):
        raise SchedulingError(
            f"Expected exactly at least {n + 1} loops, found {len(inner_loops)}"
        )
    return inner_loops[n]


def get_inner_loop(proc, loop):
    return get_nth_inner_loop(proc, loop, 0)


def is_binop(proc, expr, op=None):
    expr = proc.forward(expr)
    return isinstance(expr, BinaryOpCursor) and (op is None or expr.op() == op)


def is_add(proc, expr):
    return is_binop(proc, expr, "+")


def is_sub(proc, expr):
    return is_binop(proc, expr, "-")


def is_mul(proc, expr):
    return is_binop(proc, expr, "*")


def is_div(proc, expr):
    return is_binop(proc, expr, "/")


def is_mod(proc, expr):
    return is_binop(proc, expr, "%")


def is_builtin(proc, expr, name):
    expr = proc.forward(expr)
    return isinstance(expr, BuiltInFunctionCursor) and expr.name() == name


def is_select(proc, expr):
    return is_builtin(proc, expr, "select")


def is_relu(proc, expr):
    return is_builtin(proc, expr, "relu")


def is_sin(proc, expr):
    return is_builtin(proc, expr, "sin")


def is_literal(proc, expr, value=None):
    expr = proc.forward(expr)
    return isinstance(expr, LiteralCursor) and (value is None or expr.value() == value)


def is_reduce(proc, reduce):
    reduce = proc.forward(reduce)
    return isinstance(reduce, ReduceCursor)


def is_assign(proc, assign):
    assign = proc.forward(assign)
    return isinstance(assign, AssignCursor)


def is_read(proc, read, name=None):
    read = proc.forward(read)
    return isinstance(read, ReadCursor) and (name is None or read.name() == name)


def is_write(proc, write):
    return is_reduce(proc, write) or is_assign(proc, write)


def is_access(proc, access):
    return is_read(proc, access) or is_write(proc, access)


def is_unary_minus(proc, expr):
    expr = proc.forward(expr)
    return isinstance(expr, UnaryMinusCursor)


def is_start_of_body(proc, stmt):
    stmt = proc.forward(stmt)
    return isinstance(stmt.prev(), InvalidCursor)


def is_end_of_body(proc, stmt):
    stmt = proc.forward(stmt)
    return isinstance(stmt.next(), InvalidCursor)


def get_depth(proc, cursor):
    cursor = proc.forward(cursor)

    depth = 1
    while not isinstance(cursor, InvalidCursor):
        cursor = cursor.parent()
        depth += 1
    return depth


def get_lca(proc, cursor1, cursor2):
    """
    Gets the lowest common ancestor of [cursor1] and [cursor2].
    """
    cursor1 = proc.forward(cursor1)
    cursor2 = proc.forward(cursor2)

    c = cursor1
    while not c.is_ancestor_of(cursor2):
        c = c.parent()
        if isinstance(c, InvalidCursor):
            raise CursorNavigationError("these cursors do not have a common ancestor")
    return c


def get_distance(proc, cursor1, cursor2):
    lca = get_lca(proc, cursor1, cursor2)
    return (
        get_depth(proc, cursor1) + get_depth(proc, cursor2) - 2 * get_depth(proc, lca)
    )


def are_exprs_equal(proc, expr1, expr2):
    expr1 = proc.forward(expr1)
    expr2 = proc.forward(expr2)

    def check(expr1, expr2):
        if type(expr1) != type(expr2):
            return False

        attrs = ["name", "value", "op", "config", "field"]
        for attr in attrs:
            if (
                hasattr(expr1, attr)
                and getattr(expr1, attr)() != getattr(expr2, attr)()
            ):
                return False

        expr1_children = list(get_children(proc, expr1))
        expr2_children = list(get_children(proc, expr2))
        if len(expr1_children) != len(expr2_children):
            return False
        for c1, c2 in zip(expr1_children, expr2_children):
            if not check(c1, c2):
                return False
        return True

    return check(expr1, expr2)


def is_copy(proc, assign):
    assign = proc.forward(assign)
    if not is_assign(proc, assign):
        return False
    if not is_read(proc, assign.rhs()):
        return False
    lhs_decl = get_declaration(proc, assign, assign.name())
    rhs_decl = get_declaration(proc, assign, assign.rhs().name())
    return lhs_decl.mem() is rhs_decl.mem()


def get_bounding_block(proc, cursors):
    cursors = map(lambda c: proc.forward(c), cursors)
    cursors = list(map(lambda c: get_my_stmt(proc, c), cursors))
    depth = list(map(lambda c: get_depth(proc, c), cursors))

    def has_same_parent(cursors):
        parent = cursors[0].parent()
        return all(parent == c.parent() for c in cursors)

    while not has_same_parent(cursors):
        mx = max(depth)
        for i, c in enumerate(cursors):
            if depth[i] == mx:
                cursors[i] = c.parent()
                depth[i] -= 1
    cursors = list(map(lambda c: (c, get_index_in_body(proc, c)), cursors))
    cursors = sorted(cursors, key=lambda k: k[1])
    diff = cursors[-1][1] - cursors[0][1]
    block = cursors[0][0].as_block().expand(0, diff)
    return block


def expr_to_string(expr_cursor, subst={}):
    def expr_list_to_string(expr_list, subst):
        expr_str_list = [expr_to_string(i, subst) for i in expr_list]
        if not expr_str_list:
            return ""
        return "[" + ", ".join(expr_str_list) + "]"

    if isinstance(expr_cursor, ExprListCursor):
        return expr_list_to_string(expr_cursor, subst)

    if not isinstance(expr_cursor, ExprCursor):
        raise SchedulingError("Cursor must be an ExprCursor")
    if isinstance(expr_cursor, ReadCursor):
        name = str(expr_cursor.name())
        if name in subst:
            return f"({subst[name]})"
        idx_str = expr_list_to_string(expr_cursor.idx(), subst)
        return f"({name}{idx_str})"
    elif isinstance(expr_cursor, ReadConfigCursor):
        raise SchedulingError("ReadConfigCursor is not supported")
    elif isinstance(expr_cursor, LiteralCursor):
        val_str = str(expr_cursor.value())
        return f"({val_str})"
    elif isinstance(expr_cursor, UnaryMinusCursor):
        arg_str = expr_to_string(expr_cursor.arg, subst)
        return f"(-{arg_str})"
    elif isinstance(expr_cursor, BinaryOpCursor):
        binop_str = expr_cursor.op()
        lhs_str = expr_to_string(expr_cursor.lhs(), subst)
        rhs_str = expr_to_string(expr_cursor.rhs(), subst)
        return f"({lhs_str}{binop_str}{rhs_str})"
    elif isinstance(expr_cursor, BuiltInFunctionCursor):
        name = expr_cursor.name()
        args_str = expr_list_to_string(expr_cursor.args(), subst)
        return f"({name}({args_str[1:-1]}))"
    elif isinstance(expr_cursor, WindowExprCursor):
        raise SchedulingError("WindowExprCursor is not supported")
    else:
        assert False, "Undefined Type"



from __future__ import annotations
from typing import Tuple

from exo.API_cursors import *
from exo.range_analysis import IndexRange

from .inspection import get_parents


def index_range_analysis(expr: ExprCursor, env: dict) -> IndexRange | int:
    """
    User-level implementation of range analysis implemented inside compiler. This
    implementation is more a proof of concept, since we could really just write
    a wrapper around the compiler's range analysis.
    """
    assert isinstance(expr, ExprCursor)
    assert isinstance(env, dict)

    def analyze_range(expr) -> IndexRange | int:
        assert isinstance(expr, ExprCursor)

        if isinstance(expr, ReadCursor):
            sym = expr.name()
            if sym not in env:
                return IndexRange(expr._impl._node, 0, 0)
            lo, hi = env[sym]
            return IndexRange.create_constant_range(lo, hi)
        elif isinstance(expr, LiteralCursor):
            return expr.value()
        elif isinstance(expr, UnaryMinusCursor):
            return -analyze_range(expr.arg())
        elif isinstance(expr, BinaryOpCursor):
            lhs_range = analyze_range(expr.lhs())
            rhs_range = analyze_range(expr.rhs())
            op = expr.op()
            if op == "+":
                return lhs_range + rhs_range
            elif op == "-":
                return lhs_range - rhs_range
            elif op == "*":
                return lhs_range * rhs_range
            elif op == "/":
                return lhs_range // rhs_range
            elif op == "%":
                return lhs_range % rhs_range
            else:
                assert False, "invalid binop in index expression"
        else:
            assert False, "invalid expr in index expression"

    return analyze_range(expr)


def constant_bound(expr: ExprCursor, env: dict) -> Tuple[int, int] | None:
    """
    Returns constant integer bounds for [expr], if possible, and
    None otherwise. The bounds are inclusive.

    This is an exact copy and paste of the constant_bound function
    in the compiler's range_analysis. The only difference is that
    it calls the user-level index_range_analysis defined above.
    """
    if isinstance(expr, int):
        return (expr, expr)

    idx_rng = index_range_analysis(expr, env)
    if isinstance(idx_rng, int):
        return (idx_rng, idx_rng)

    if idx_rng.base is not None:
        return (None, None)
    return (idx_rng.lo, idx_rng.hi)


def infer_range(idx_expr: Cursor, scope: Cursor) -> IndexRange:
    """
    Infers the range of possible values of [idx_expr] within [scope].
    """
    assert isinstance(idx_expr, Cursor)
    assert isinstance(scope, Cursor)
    env = dict()

    # Only add bound variables to the env (which excludes scope)
    ancestors = list(get_parents(idx_expr.proc(), idx_expr, up_to=scope))[:-1]
    for c in filter(lambda x: isinstance(x, ForCursor), ancestors):
        lo, _ = constant_bound(c.lo(), env)
        _, hi = constant_bound(c.hi(), env)
        if hi is not None:
            hi -= 1  # loop upper bound is exclusive
        env[c.name()] = (lo, hi)

    bounds = index_range_analysis(idx_expr, env)

    if isinstance(bounds, int):
        return IndexRange.create_int(bounds, bounds)

    return bounds


def bounds_inference(proc, loop, buffer_name: str, buffer_dim: int, include=["W"]):
    """ """
    loop = proc.forward(loop)

    matches = []
    # TODO: also want probably reduces... for both read and write
    if "R" in include:
        # TODO: proc.find doesn't take a scope. Either write a variant or add that as an optional arg
        # Also, proc.find fails if no matches are found...but we really just want it to return []
        matches += proc.find(f"{buffer_name}[_]", many=True)
    if "W" in include:
        matches += proc.find(f"{buffer_name}[_] = _", many=True)

    bound = None  # None is basically bottom in abstract interpretation
    for c in matches:
        idx_expr = c.idx()[buffer_dim]
        # TODO: This implementation is slower, but easier to understand than a tree traversal
        # because we have to rebuild the environment for each infer_range call.
        cur_bounds = infer_range(idx_expr, loop)

        if bound is None:
            bound = cur_bounds
        else:
            bound |= cur_bounds
    return bound



from __future__ import annotations

import pytest

from exo import proc, SchedulingError, Procedure
from exo.LoopIR import LoopIR, T
from exo.LoopIR_pprint import _print_cursor
from exo.internal_cursors import (
    Cursor,
    Block,
    InvalidCursorError,
    Node,
)
from exo.pattern_match import match_pattern
from exo.prelude import Sym
from exo.syntax import size, f32


def _find_cursors(ctx, pattern):
    if isinstance(ctx, Procedure):
        ctx = ctx._root()

    if isinstance(ctx, LoopIR.proc):
        ctx = Cursor.create(ctx)

    cursors = match_pattern(ctx, pattern, call_depth=1)
    assert isinstance(cursors, list)
    if not cursors:
        raise SchedulingError("failed to find matches", pattern=pattern)
    return cursors


def _find_stmt(ctx, pattern):
    curs = _find_cursors(ctx, pattern)
    assert len(curs) == 1
    curs = curs[0]
    if len(curs) != 1:
        raise SchedulingError(
            "pattern did not match a single statement", pattern=pattern
        )
    return curs[0]


@pytest.fixture(scope="session")
def proc_foo():
    @proc
    def foo(n: size, m: size):
        for i in seq(0, n):
            for j in seq(0, m):
                x: f32
                x = 0.0
                y: f32
                y = 1.1

    yield foo


@pytest.fixture(scope="session")
def proc_bar():
    @proc
    def bar(n: size, m: size):
        x: f32
        for i in seq(0, n):
            for j in seq(0, m):
                x = 0.0
                x = 1.0
                x = 2.0
                x = 3.0
                x = 4.0
                x = 5.0

    yield bar


@pytest.fixture(scope="session")
def proc_baz():
    @proc
    def baz(n: size, m: size):
        for i in seq(0, n):
            for j in seq(0, m):
                x: f32
                x = 0.0
                y: f32
                y = 1.1
                for k in seq(0, n):
                    pass
                    pass

    yield baz


def test_get_root(proc_foo):
    cursor = proc_foo._root()
    assert cursor._node is proc_foo.INTERNAL_proc()


def test_find_cursor(proc_foo):
    c = _find_cursors(proc_foo, "for j in _:_")
    assert len(c) == 1
    c = c[0]  # One match
    c = c[0]  # First/only node in the block

    assert c._node is proc_foo.INTERNAL_proc().body[0].body[0]


def test_find_hole_in_middle(proc_bar):
    c_body_1_5 = _find_cursors(proc_bar, "x = 1.0 ; _ ; x = 4.0")[0]
    for_j = _find_stmt(proc_bar, "for j in _: _")
    assert c_body_1_5 == for_j.body()[1:5]


def test_gap_insert_pass(proc_foo, golden):
    c = _find_stmt(proc_foo, "x = 0.0")
    assn = c._node
    g = c.after()
    foo2, _ = g._insert([LoopIR.Pass(None, assn.srcinfo)])
    assert str(foo2) == golden


def test_insert_root_front(proc_foo, golden):
    c = proc_foo._root()
    foo2, _ = c.body().before()._insert([LoopIR.Pass(None, c._node.srcinfo)])
    assert str(foo2) == golden


def test_insert_root_end(proc_foo, golden):
    c = proc_foo._root()
    foo2, _ = c.body().after()._insert([LoopIR.Pass(None, c._node.srcinfo)])
    assert str(foo2) == golden


def test_block_gaps(proc_bar):
    c = _find_stmt(proc_bar, "for j in _: _")

    body = c.body()
    assert len(body) == 6
    subset = body[1:4]
    assert len(subset) == 3

    cx1 = _find_stmt(proc_bar, "x = 1.0")
    cx3 = _find_stmt(proc_bar, "x = 3.0")
    assert subset[0] == cx1
    assert subset[2] == cx3

    assert subset.before() == cx1.before()
    assert subset.after() == cx3.after()


def test_block_sequence_interface(proc_bar):
    for_j_body = _find_stmt(proc_bar, "for j in _: _").body()
    assert for_j_body[:] is not for_j_body
    assert for_j_body[:] == for_j_body
    assert len(list(for_j_body)) == 6

    with pytest.raises(IndexError, match="block cursors must be contiguous"):
        # noinspection PyStatementEffect
        # Sequence's __getitem__ should throw here, so this does have a side effect
        for_j_body[::2]


def test_block_delete(proc_bar, golden):
    c = _find_stmt(proc_bar, "for j in _: _")
    stmts = c.body()[1:4]

    bar2, _ = stmts._delete()
    assert str(bar2) == golden


def test_block_replace(proc_bar, golden):
    c = _find_stmt(proc_bar, "for j in _: _")
    stmts = c.body()[1:4]

    bar2, _ = stmts._replace([LoopIR.Pass(None, c._node.srcinfo)])
    assert str(bar2) == golden


def test_block_delete_whole_block(proc_bar, golden):
    c = _find_stmt(proc_bar, "for j in _: _")
    bar2, _ = c.body()._delete()
    assert str(bar2) == golden


def test_node_replace(proc_bar, golden):
    c = _find_stmt(proc_bar, "x = 3.0")
    assert isinstance(c, Node)

    bar2, _ = c._replace([LoopIR.Pass(None, c._node.srcinfo)])
    assert str(bar2) == golden


def test_cursor_move(proc_foo):
    c = _find_stmt(proc_foo, "for j in _:_")

    c_list = c.body()  # list of j's body
    assert isinstance(c_list, Block)
    assert len(c_list) == 4
    assert c_list.parent() == c

    c1 = c_list[0]  # x : f32
    assert c1.parent() == c
    assert c1._node is proc_foo.INTERNAL_proc().body[0].body[0].body[0]

    c2 = c1.next()  # x = 0.0
    assert c2.parent() == c
    assert c2._node is proc_foo.INTERNAL_proc().body[0].body[0].body[1]

    c3 = c1.next(2)  # y : f32
    assert c3.parent() == c
    assert c3._node is proc_foo.INTERNAL_proc().body[0].body[0].body[2]

    _c2_ = c3.prev()
    assert c2 == _c2_
    _c1_ = c3.prev(2)
    assert c1 == _c1_


def test_cursor_move_invalid(proc_foo):
    # Edge cases near the root
    c = proc_foo._root()
    with pytest.raises(InvalidCursorError, match="cannot move root cursor"):
        c.next()

    with pytest.raises(InvalidCursorError, match="cursor does not have a parent"):
        c.parent()

    # Edge cases near expressions
    c = _find_cursors(proc_foo, "m")[0]
    with pytest.raises(InvalidCursorError, match="cursor is not inside block"):
        c.next()

    # Edge cases near first statement in block
    c = _find_stmt(proc_foo, "x: f32")
    with pytest.raises(InvalidCursorError, match="cursor is out of range"):
        c.prev()

    c.before()  # ok
    with pytest.raises(InvalidCursorError, match="cursor is out of range"):
        c.prev()

    # Edge cases near last statement in block
    c = _find_stmt(proc_foo, "y = 1.1")

    c.after()  # ok
    with pytest.raises(InvalidCursorError, match="cursor is out of range"):
        c.next()


def test_cursor_gap(proc_foo):
    # for i in seq(0, n):
    #    for j in seq(0, m):
    #        x: f32   <- x_alloc
    #                 <- g1
    #        x = 0.0  <- x_assn
    #        y: f32
    #                 <- g2
    #        y = 1.1  <- y_assn
    for_j = _find_stmt(proc_foo, "for j in _:_")
    x_alloc = _find_stmt(proc_foo, "x: f32")
    x_assn = _find_stmt(proc_foo, "x = 0.0")
    y_assn = _find_stmt(proc_foo, "y = 1.1")

    assert str(x_alloc._node) == "x: f32 @ DRAM"

    g1 = x_alloc.after()
    assert g1.anchor() == x_alloc
    assert g1.parent() == for_j

    g2 = g1.anchor().next(2).after()
    assert g2 == x_assn.next().after()
    assert g2.anchor().prev() == x_assn
    assert g2.anchor().next() == y_assn
    assert g2.parent() == for_j


def test_cursor_replace_expr(proc_foo, golden):
    c = _find_cursors(proc_foo, "m")[0]
    foo2, _ = c._replace(LoopIR.Const(42, T.size, c._node.srcinfo))
    assert str(foo2) == golden


def test_cursor_cannot_convert_expr_to_block(proc_foo):
    c = _find_cursors(proc_foo, "m")[0]
    with pytest.raises(InvalidCursorError, match="node is not inside a block"):
        c.as_block()


def test_cursor_replace_expr_deep(golden):
    @proc
    def example():
        x: f32
        x = 1.0 * (2.0 + 3.0)
        # x = 1.0 * (4.0 + 3.0)

    c: Node = _find_cursors(example, "2.0")[0]
    four = LoopIR.Const(4.0, T.f32, c._node.srcinfo)

    example_new, _ = c._replace(four)
    assert str(example_new) == golden


def test_cursor_forward_expr_deep():
    @proc
    def example():
        x: f32
        x = 1.0 * (2.0 + 3.0)
        # x = 1.0 * (4.0 + 3.0)

    c2: Node = _find_cursors(example, "2.0")[0]
    four = LoopIR.Const(4.0, T.f32, c2._node.srcinfo)

    example_new, fwd = c2._replace(four)
    assert fwd(c2) == _find_cursors(example_new, "4.0")[0]

    assert fwd(_find_stmt(example, "x = _")) == _find_stmt(example_new, "x = _")
    assert (
        fwd(_find_cursors(example, "_ + _")[0])
        == _find_cursors(example_new, "_ + _")[0]
    )
    assert fwd(_find_cursors(example, "1.0")[0]) == _find_cursors(example_new, "1.0")[0]
    assert fwd(_find_cursors(example, "3.0")[0]) == _find_cursors(example_new, "3.0")[0]


def test_cursor_loop_bound(proc_foo):
    c_for_i = proc_foo._root().body()[0]
    c_bound = c_for_i._child_node("hi")
    assert isinstance(c_bound._node, LoopIR.Read)


def test_cursor_invalid_child(proc_foo):
    c = proc_foo._root()

    # Quick sanity check
    assert c._child_node("body", 0) == c.body()[0]

    with pytest.raises(AttributeError, match="has no attribute '_invalid'"):
        c._child_node("_invalid", None)

    with pytest.raises(ValueError, match="must index into block attribute"):
        c._child_node("body", None)

    with pytest.raises(InvalidCursorError, match="cursor is out of range"):
        c._child_node("body", 42)


def test_insert_forward_orelse():
    @proc
    def example_old():
        x: f32
        if 1 < 2:
            x = 1.0
        else:
            x = 2.0

    x1_old = _find_stmt(example_old, "x = 1.0")
    x2_old = _find_stmt(example_old, "x = 2.0")
    gap = x1_old.after()

    stmt = [LoopIR.Pass(None, x1_old._node.srcinfo)]

    example_new, fwd = gap._insert(stmt)
    x2_new = _find_stmt(example_new, "x = 2.0")

    assert fwd(x2_old) == x2_new


def test_double_insert_forwarding(golden):
    @proc
    def proc_s1():
        x: f32
        if 1 < 2:
            x = 1.0
            # x = 2.0  (added in s2)
        else:
            x = 3.0
            # x = 4.0  (added in s3)

    x1_s1 = _find_stmt(proc_s1, "x = 1.0")
    x3_s1 = _find_stmt(proc_s1, "x = 3.0")

    x2_stmt = x1_s1._node.update(rhs=LoopIR.Const(2.0, T.f32, x1_s1._node.srcinfo))
    x4_stmt = x1_s1._node.update(rhs=LoopIR.Const(4.0, T.f32, x1_s1._node.srcinfo))

    proc_s2, fwd_12 = x1_s1.after()._insert([x2_stmt])
    proc_s3, fwd_23 = fwd_12(x3_s1).after()._insert([x4_stmt])
    assert str(proc_s3) == golden

    def fwd_13(cur):
        return fwd_23(fwd_12(cur))

    x1_s3 = _find_stmt(proc_s3, "x = 1.0")
    assert fwd_13(x1_s1) == x1_s3

    x3_s3 = _find_stmt(proc_s3, "x = 3.0")
    assert fwd_13(x3_s1) == x3_s3

    if_pat = "if _: _\nelse: _"
    assert fwd_13(_find_stmt(proc_s1, if_pat)) == _find_stmt(proc_s3, if_pat)

    with pytest.raises(InvalidCursorError, match="cannot forward from unknown root"):
        fwd_23(x1_s1)


@pytest.mark.parametrize(
    "old, new",
    [
        (0, 0),
        (1, 1),
        (2, None),
        (3, None),
        (4, 2),
        (5, 3),
    ],
)
def test_delete_forward_node(proc_bar, old, new):
    for_j = _find_stmt(proc_bar, "for j in _: _").body()
    bar_new, fwd = for_j[2:4]._delete()
    for_j_new = _find_stmt(bar_new, "for j in _: _").body()

    if new is None:
        with pytest.raises(InvalidCursorError, match="node no longer exists"):
            fwd(for_j[old])
    else:
        assert fwd(for_j[old]) == for_j_new[new]


@pytest.mark.parametrize(
    "old, new",
    [
        ("x = 0.0", "x = 0.0"),
        ("x = 1.0", None),
        ("x = 2.0", None),
        ("x = 3.0", None),
        ("x = 4.0", "x = 4.0"),
        ("x = 5.0", "x = 5.0"),
    ],
)
def test_block_replace_forward_node(proc_bar, old, new):
    for_j = _find_stmt(proc_bar, "for j in _: _").body()
    bar_new, fwd = for_j[1:4]._replace(
        [
            LoopIR.Pass(None, for_j.parent()._node.srcinfo),
            LoopIR.Pass(None, for_j.parent()._node.srcinfo),
        ]
    )

    old_c = _find_stmt(proc_bar, old)

    if new is None:
        with pytest.raises(InvalidCursorError, match="node no longer exists"):
            fwd(old_c)
    else:
        bar_new = Cursor.create(bar_new)
        assert fwd(old_c) == match_pattern(bar_new, new)[0][0]


def test_cursor_pretty_print_nodes(proc_bar, golden):
    output = []

    ir = proc_bar._loopir_proc

    root = Cursor.create(ir)
    output.append(_print_cursor(root))

    c = _find_stmt(ir, "for i in _: _")
    output.append(_print_cursor(c))

    c = _find_stmt(ir, "for j in _: _")
    output.append(_print_cursor(c))

    c = _find_stmt(ir, "x = 0.0")
    output.append(_print_cursor(c))

    c = _find_stmt(ir, "x = 2.0")
    output.append(_print_cursor(c))

    assert "\n\n".join(output) == golden


def test_cursor_pretty_print_gaps(proc_bar, golden):
    output = []

    c = _find_stmt(proc_bar, "x: f32").before()
    output.append(_print_cursor(c))

    c = _find_stmt(proc_bar, "for i in _: _").before()
    output.append(_print_cursor(c))

    c = _find_stmt(proc_bar, "for j in _: _").before()
    output.append(_print_cursor(c))

    c = _find_stmt(proc_bar, "x = 0.0").before()
    output.append(_print_cursor(c))

    c = _find_stmt(proc_bar, "x = 2.0").before()
    output.append(_print_cursor(c))

    c = _find_stmt(proc_bar, "x = 5.0").after()
    output.append(_print_cursor(c))

    assert "\n\n".join(output) == golden


def test_cursor_pretty_print_blocks(proc_bar, golden):
    output = []

    c = _find_stmt(proc_bar, "for j in _: _").as_block()
    output.append(_print_cursor(c))

    c = _find_cursors(proc_bar, "x = 1.0; _; x = 3.0")[0]
    output.append(_print_cursor(c))

    c = _find_cursors(proc_bar, "x = 0.0; x = 1.0")[0]
    output.append(_print_cursor(c))

    c = _find_cursors(proc_bar, "x = 4.0; x = 5.0")[0]
    output.append(_print_cursor(c))

    c = _find_cursors(proc_bar, "x = 0.0; _; x = 5.0")[0]
    output.append(_print_cursor(c))

    assert "\n\n".join(output) == golden


def test_move_block(proc_bar, golden):
    c = _find_cursors(proc_bar, "x = 1.0 ; x = 2.0")[0]

    # Movement within block
    p0, _ = c._move(c[0].prev().before())
    p1, _ = c._move(c.before())
    p2, _ = c._move(c.after())
    p3, _ = c._move(c[-1].next().after())
    p4, _ = c._move(c[-1].next(2).after())
    p5, _ = c._move(c[-1].next(3).after())

    assert str(p1) == str(p2), "Both before and after should keep block in place."

    # Movement upward
    pu0, _ = c._move(c.parent().before())
    pu1, _ = c._move(c.parent().after())
    pu2, _ = c._move(c.parent().parent().before())
    pu3, _ = c._move(c.parent().parent().prev().before())
    pu4, _ = c._move(c.parent().parent().after())

    # Movement downward (abbreviated)
    c2 = _find_cursors(proc_bar, "x: _")[0]
    pd0, _ = c2._move(c[0].prev().before())
    pd1, _ = c2._move(c.before())
    pd2, _ = c2._move(c2[-1].next().after())

    # Move out a whole loop (needs to insert pass)
    c3 = _find_cursors(proc_bar, "for j in _: _")[0]
    pl0, _ = c3._move(c3.parent().after())

    all_tests = [p0, p1, p2, p3, p4, p5, pu0, pu1, pu2, pu3, pu4, pd0, pd1, pd2, pl0]
    actual = "\n".join(str(p) for p in all_tests)
    assert actual == golden


def _debug_print_forwarding(x, fwd):
    fx = fwd(x)
    print(x._path)
    print(_print_cursor(x))
    print(x._root)
    print("-----------")
    print(fx._path)
    print(_print_cursor(fx))
    print(fx._root)
    print()
    print()


def test_move_block_forwarding(proc_bar, golden):
    c = _find_cursors(proc_bar, "x = 1.0 ; x = 2.0")[0]

    x_orig = []
    for i in range(6):
        x_orig.append(_find_stmt(proc_bar, f"x = {i}.0"))

    def _test_fwd(fwd):
        for x in x_orig:
            # _debug_print_forwarding(x, fwd)
            assert str(fwd(x)._node) == str(x._node)

    # Movement within block
    _, fwd0 = c._move(c[0].prev().before())
    _test_fwd(fwd0)
    _, fwd1 = c._move(c.before())
    _test_fwd(fwd1)
    _, fwd2 = c._move(c.after())
    _test_fwd(fwd2)
    _, fwd3 = c._move(c[-1].next().after())
    _test_fwd(fwd3)
    _, fwd4 = c._move(c[-1].next(2).after())
    _test_fwd(fwd4)
    _, fwd5 = c._move(c[-1].next(3).after())
    _test_fwd(fwd5)

    # Movement upward
    _, fwd6 = c._move(c.parent().before())
    _test_fwd(fwd6)
    _, fwd7 = c._move(c.parent().after())
    _test_fwd(fwd7)
    _, fwd8 = c._move(c.parent().parent().before())
    _test_fwd(fwd8)
    _, fwd9 = c._move(c.parent().parent().prev().before())
    _test_fwd(fwd9)
    _, fwd10 = c._move(c.parent().parent().after())
    _test_fwd(fwd10)

    # Movement downward (abbreviated)
    c2 = _find_cursors(proc_bar, "x: _")[0]
    _, fwd11 = c2._move(c[0].prev().before())
    _test_fwd(fwd11)
    _, fwd12 = c2._move(c.before())
    _test_fwd(fwd12)
    _, fwd13 = c2._move(c2[-1].next().after())
    _test_fwd(fwd13)

    # Move out a whole loop
    c3 = _find_cursors(proc_bar, "for j in _: _")[0]
    _, fwd14 = c3._move(c3.parent().after())
    _test_fwd(fwd14)


def test_wrap_block(proc_bar, golden):
    k = Sym("k")

    def wrapper(body):
        src = body[0].srcinfo
        zero = LoopIR.Const(0, T.index, src)
        eight = LoopIR.Const(8, T.index, src)
        return LoopIR.For(k, zero, eight, body, LoopIR.Seq(), None, src)

    procs = []
    for i in range(0, 6):
        for j in range(i + 1, 6):
            c = _find_cursors(proc_bar, f"x = {i}.0 ; _ ; x = {j}.0")[0]
            p, _ = c._wrap(wrapper, "body")
            procs.append(p)

    assert "\n".join(map(str, procs)) == golden


def test_wrap_block_forward(proc_bar):
    x_orig = []
    for i in range(6):
        x_orig.append(_find_stmt(proc_bar, f"x = {i}.0"))

    def _test_fwd(fwd):
        for x in x_orig:
            # _debug_print_forwarding(x, fwd)
            assert str(fwd(x)._node) == str(x._node)

    def wrapper(orelse):
        src = orelse[0].srcinfo
        true = LoopIR.Const(True, T.bool, src)
        return LoopIR.If(true, [LoopIR.Pass(None, src)], orelse, None, src)

    for i in range(0, 6):
        for j in range(i + 1, 6):
            c = _find_cursors(proc_bar, f"x = {i}.0 ; _ ; x = {j}.0")[0]
            _, fwd = c._wrap(wrapper, "orelse")
            _test_fwd(fwd)


def test_move_forward_diff_scopes_1(golden):
    @proc
    def foo():
        x: i8
        y: i8
        z: i8
        for i in seq(0, 4):
            pass

    alloc_xy = foo.find("x: _")._impl.as_block().expand(0, 1)
    pass_c = foo.find("pass")._impl
    ir, fwd = alloc_xy._move(pass_c.before())
    assert fwd(alloc_xy[0])._path == [("body", 1), ("body", 0)]
    assert str(ir) == golden

    @proc
    def foo():
        z: i8
        for i in seq(0, 4):
            pass
        x: i8
        y: i8

    alloc_xy = foo.find("x: _")._impl.as_block().expand(0, 1)
    pass_c = foo.find("pass")._impl
    ir, fwd = alloc_xy._move(pass_c.before())
    assert fwd(alloc_xy[0])._path == [("body", 1), ("body", 0)]
    assert str(ir) == golden


def test_move_forward_diff_scopes_2():
    @proc
    def foo():
        pass
        for i in seq(0, 4):
            x: i8
            y: i8
            z: i8

    alloc_xy = foo.find("x: _")._impl.as_block().expand(0, 1)
    pass_c = foo.find("pass")._impl
    ir, fwd = alloc_xy._move(pass_c.after())
    assert fwd(alloc_xy[0])._path == [("body", 1)]
    assert fwd(alloc_xy[1])._path == [("body", 2)]

    @proc
    def foo():
        for i in seq(0, 4):
            x: i8
            y: i8
            z: i8
        pass

    alloc_xy = foo.find("x: _")._impl.as_block().expand(0, 1)
    pass_c = foo.find("pass")._impl
    ir, fwd = alloc_xy._move(pass_c.after())
    assert fwd(alloc_xy[0])._path == [("body", 2)]
    assert fwd(alloc_xy[1])._path == [("body", 3)]


def test_move_forward_if_orelse(golden):
    @proc
    def foo():
        if True:
            x: i8
        else:
            y: i8

    alloc_x = foo.find("x: _")._impl
    alloc_y = foo.find("y: _")._impl
    ir, fwd = alloc_x._move(alloc_y.after())
    assert fwd(alloc_x)._path == [("body", 0), ("orelse", 1)]
    assert str(ir) == golden


def test_insert_forwarding_for_blocks(proc_baz, golden):
    c = _find_stmt(proc_baz, "x = 0.0")

    b_above = c.parent()
    b_edit = b_above.body()
    b_below = b_edit[-1].body()
    b_before = b_edit[:2]
    b_after = b_edit[2:]

    _, fwd = c.after()._insert([LoopIR.Pass(None, c._node.srcinfo)])

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level
    output.append(_print_cursor(fwd(b_below)))  # below edit level
    output.append(_print_cursor(fwd(b_before)))  # same edit level, disjoint
    output.append(_print_cursor(fwd(b_after)))  # same dit level, disjoint

    # Blocks containing the insertion point also work intuitively.
    output.append(_print_cursor(fwd(b_edit)))

    assert "\n\n".join(output) == golden


def test_delete_forwarding_for_blocks(proc_baz, golden):
    c = _find_cursors(proc_baz, "x = 0.0; y: _")[0]

    b_above = c.parent()
    b_edit = b_above.body()
    b_below = b_edit[-1].body()
    b_before = b_edit[:1]
    b_after = b_edit[3:]

    b_aligned_with_deletion_on_left = b_edit[1:]
    b_aligned_with_deletion_on_right = b_edit[:3]
    b_with_endpoint_in_deletion = b_edit[:2]

    _, fwd = c._delete()

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level
    output.append(_print_cursor(fwd(b_below)))  # below edit level
    output.append(_print_cursor(fwd(b_before)))  # same edit level, disjoint
    output.append(_print_cursor(fwd(b_after)))  # same edit level, disjoint

    # Blocks entirely containing the deletion block also work intuitively.
    output.append(_print_cursor(fwd(b_edit)))
    output.append(_print_cursor(fwd(b_aligned_with_deletion_on_left)))
    output.append(_print_cursor(fwd(b_aligned_with_deletion_on_right)))

    # Blocks partially containing the deletion block don't work.
    with pytest.raises(InvalidCursorError, match=r"block no longer exists"):
        fwd(b_with_endpoint_in_deletion)

    assert "\n\n".join(output) == golden


def test_delete_forwarding_for_blocks_fail():
    @proc
    def foo():
        for i in seq(0, 4):
            x: i8
            y: i8
            z: i8

    body = _find_stmt(foo, "for i in _: _").body()
    _, fwd = body.parent()._delete()

    with pytest.raises(
        InvalidCursorError, match=r"block no longer exists \(parent deleted\)"
    ):
        fwd(body)


def test_block_replace_forwarding_for_blocks(proc_baz, golden):
    c = _find_cursors(proc_baz, "x = 0.0; _; y = 1.1")[0]

    b_above = c.parent()
    b_edit = b_above.body()
    b_below = b_edit[-1].body()
    b_before = b_edit[:1]
    b_after = b_edit[4:]
    b_with_endpoint_in_replace = b_edit[:2]
    b_inside = b_edit[2:4]

    # replace 1:4 with two pass stmts
    pass_ir = LoopIR.Pass(None, c[0]._node.srcinfo)
    _, fwd = c._replace([pass_ir, pass_ir])

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level
    output.append(_print_cursor(fwd(b_below)))  # below edit level
    output.append(_print_cursor(fwd(b_before)))  # same edit level, disjoint
    output.append(_print_cursor(fwd(b_after)))  # same edit level, disjoint

    # Blocks containing the entire replace block work intuitively.
    output.append(_print_cursor(fwd(b_edit)))
    output.append(_print_cursor(fwd(c)))

    # Blocks partially containing the replace block are invalidated.
    with pytest.raises(InvalidCursorError, match=r"block no longer exists"):
        fwd(b_with_endpoint_in_replace)

    # Blocks entirely within the replace block are invalidated
    with pytest.raises(InvalidCursorError, match=r"block no longer exists"):
        fwd(b_inside)

    assert "\n\n".join(output) == golden


def test_node_replace_forwarding(proc_baz):
    c = _find_cursors(proc_baz, "1.1")[0]
    _, fwd = c._replace(LoopIR.Const(42.0, T.f32, c._node.srcinfo))
    assert fwd(c)._node.val == 42.0


def test_block_replace_forwarding_stmt_to_stmt(proc_baz):
    c = _find_stmt(proc_baz, "x = 0")
    _, fwd = c._replace(LoopIR.Pass(None, c._node.srcinfo))
    assert isinstance(fwd(c)._node, LoopIR.Pass)


def test_wrap_forwarding_for_blocks(proc_baz, golden):
    c = _find_cursors(proc_baz, "y: _; y = 1.1")[0]

    b_above = c.parent()
    b_edit = b_above.body()
    b_below = b_edit[-1].body()

    b_before = b_edit[:2]
    b_after = b_edit[4:]
    b_with_endpoint_in_replace = b_edit[:3]

    # wrap 2:4 with a for loop
    k = Sym("k")

    def wrapper(body):
        src = body[0].srcinfo
        zero = LoopIR.Const(0, T.index, src)
        eight = LoopIR.Const(8, T.index, src)
        return LoopIR.For(k, zero, eight, body, LoopIR.Seq(), None, src)

    _, fwd = c._wrap(wrapper, "body")

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level
    output.append(_print_cursor(fwd(b_below)))  # below edit level
    output.append(_print_cursor(fwd(b_before)))  # same edit level, disjoint
    output.append(_print_cursor(fwd(b_after)))  # same edit level, disjoint

    # Blocks containing the entire wrap block work intuitively.
    output.append(_print_cursor(fwd(b_edit)))

    # Blocks within the wrap block work intuitively
    output.append(_print_cursor(fwd(c)))

    # Blocks partially containing the wrap block don't work.
    with pytest.raises(InvalidCursorError, match=r"block no longer exists"):
        output.append(_print_cursor(fwd(b_with_endpoint_in_replace)))

    assert "\n\n".join(output) == golden


def test_move_forwarding_for_blocks(proc_baz, golden):
    c = _find_cursors(proc_baz, "y: _; y = 1.1")[0]
    g = c[0].prev().before()

    b_above = c.parent()
    b_edit = b_above.body()
    b_below = b_edit[-1].body()

    b_with_gap = b_edit[:2]
    b_without_gap = b_edit[4:]
    b_with_endpoint_in_moved_block = b_edit[:3]

    _, fwd = c._move(g)

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level
    output.append(_print_cursor(fwd(b_below)))  # below edit level
    output.append(_print_cursor(fwd(b_with_gap)))  # contains gap we are moving block to
    output.append(_print_cursor(fwd(b_without_gap)))

    # Blocks containing the entire moved block work intuitively.
    output.append(_print_cursor(fwd(b_edit)))

    # Blocks within the moved block work intuitively
    output.append(_print_cursor(fwd(c)))

    # Blocks partially containing the wrap block don't work.
    with pytest.raises(
        InvalidCursorError,
        match=r"move cannot forward block because exactly one endpoint",
    ):
        output.append(_print_cursor(fwd(b_with_endpoint_in_moved_block)))

    assert "\n\n".join(output) == golden


def test_move_forwarding_for_blocks_gap_after(proc_baz, golden):
    c = _find_cursors(proc_baz, "y: _; y = 1.1")[0]
    g = _find_stmt(proc_baz, "pass #0").after()

    b_above = c.parent()
    b_moved_block_body = b_above.body()
    b_gap_body = g.anchor().parent().body()

    b_with_endpoint_in_moved_block = b_moved_block_body[:3]

    _, fwd = c._move(g)

    output = []
    output.append(_print_cursor(fwd(b_above)))  # above edit level

    # Blocks containing the entire moved block work intuitively.
    output.append(_print_cursor(fwd(b_moved_block_body)))

    # Blocks containing the gap we are moving block to work intuitively
    output.append(_print_cursor(fwd(b_gap_body)))

    # Blocks within the moved block work intuitively
    output.append(_print_cursor(fwd(c)))

    # Blocks partially containing the wrap block don't work.
    with pytest.raises(
        InvalidCursorError,
        match=r"move cannot forward block because exactly one endpoint",
    ):
        output.append(_print_cursor(fwd(b_with_endpoint_in_moved_block)))

    assert "\n\n".join(output) == golden



from __future__ import annotations

import itertools

import numpy as np
import pytest

from exo import proc
from exo.platforms.x86 import *
from exo.stdlib.scheduling import *

old_split = repeat(divide_loop)
old_unroll = repeat(unroll_loop)


def old_fission_after(proc, stmt_pattern, n_lifts=1):
    def find_stmts(p):
        return [c.after() for c in p.find_all(stmt_pattern)]

    return loop_hack(autofission, find_stmts)(proc, n_lifts)


@pytest.mark.isa("AVX2")
def test_avx2_memcpy(compiler):
    """
    Compute dst = src
    """

    @proc
    def memcpy_avx2(n: size, dst: R[n] @ DRAM, src: R[n] @ DRAM):  # pragma: no cover
        for i in seq(0, (n + 7) / 8):
            if n - 8 * i >= 8:
                tmp: f32[8] @ AVX2
                mm256_loadu_ps(tmp, src[8 * i : 8 * i + 8])
                mm256_storeu_ps(dst[8 * i : 8 * i + 8], tmp)
            else:
                for j in seq(0, n - 8 * i):
                    dst[8 * i + j] = src[8 * i + j]

    # TODO: -march=skylake here is a hack. Such flags should be somehow handled
    #   automatically. Maybe this should be inferred by the use of AVX2, but
    #   "skylake" isn't right anyway. We might need a first-class notion of
    #   a Target, which has certain memories available. Then we can say that
    #   e.g. Skylake-X has AVX2, AVX512, etc.
    fn = compiler.compile(
        memcpy_avx2, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    for n in (7, 8, 9, 31, 32, 33, 127, 128, 129):
        inp = np.array([float(i) for i in range(n)], dtype=np.float32)
        out = np.array([float(0) for _ in range(n)], dtype=np.float32)
        fn(None, n, out, inp)

        assert np.array_equal(inp, out)


@pytest.mark.isa("AVX2")
def test_avx2_simple_math(compiler):
    """
    Compute x = x * y^2
    """

    @proc
    def simple_math_avx2(n: size, x: R[n] @ DRAM, y: R[n] @ DRAM):  # pragma: no cover
        assert n % 8 == 0
        for i in seq(0, n / 8):
            xVec: f32[8] @ AVX2
            tmp: f32[8] @ AVX2
            yVec: f32[8] @ AVX2
            mm256_loadu_ps(xVec, x[8 * i : 8 * i + 8])
            mm256_loadu_ps(yVec, y[8 * i : 8 * i + 8])
            mm256_mul_ps(tmp, xVec, yVec)
            mm256_mul_ps(xVec, tmp, yVec)
            mm256_storeu_ps(x[8 * i : 8 * i + 8], xVec)

    fn = compiler.compile(
        simple_math_avx2, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    for n in (8, 16, 24, 32, 64, 128):
        x = np.array([float(i) for i in range(n)], dtype=np.float32)
        y = np.array([float(3 * i) for i in range(n)], dtype=np.float32)
        expected = x * y * y

        fn(None, n, x, y)
        np.testing.assert_almost_equal(x, expected)


@pytest.fixture
def simple_math_avx2_sched():
    """
    Compute x = x * y^2
    """

    @proc
    def simple_math_avx2_sched(
        n: size, x: R[n] @ DRAM, y: R[n] @ DRAM
    ):  # pragma: no cover
        for i in seq(0, n):
            x[i] = x[i] * y[i] * y[i]

    def sched_simple_math_avx2_sched(p=simple_math_avx2_sched):
        p = old_split(p, "i", 8, ["io", "ii"], tail="cut_and_guard")

        p = stage_mem(p, "for ii in _:_", "x[8 * io: 8 * io + 8]", "xVec")
        p = set_memory(p, "xVec", AVX2)

        p = replace(p, "for i0 in _:_ #0", mm256_loadu_ps)
        p = replace(p, "for i0 in _:_ #0", mm256_storeu_ps)

        p = bind_expr(p, p.find("y[_]", many=True), "yVec")
        p = autolift_alloc(p, "yVec: _", keep_dims=True)
        p = set_memory(p, "yVec", AVX2)
        p = old_fission_after(p, "yVec[_] = _")

        p = replace_all(p, mm256_loadu_ps)

        p = bind_expr(p, "xVec[_] * yVec[_]", "xy")
        p = autolift_alloc(p, "xy: _", keep_dims=True)
        p = set_memory(p, "xy", AVX2)
        p = old_fission_after(p, "xy[_] = _")

        p = replace_all(p, mm256_mul_ps)
        p = simplify(p)
        return p

    simple_math_avx2_sched = sched_simple_math_avx2_sched()

    return simple_math_avx2_sched


def test_gen_avx2_simple_math_scheduling(golden, simple_math_avx2_sched):
    assert str(simple_math_avx2_sched) == golden


@pytest.mark.isa("AVX2")
def test_exec_avx2_simple_math_scheduling(compiler, simple_math_avx2_sched):
    fn = compiler.compile(
        simple_math_avx2_sched, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    for n in (8, 16, 24, 32, 64, 128):
        x = np.array([float(i) for i in range(n)], dtype=np.float32)
        y = np.array([float(3 * i) for i in range(n)], dtype=np.float32)
        expected = x * y * y

        fn(None, n, x, y)
        np.testing.assert_almost_equal(x, expected)


def _sgemm_test_cases(fn, M, N, K):
    for m, n, k in itertools.product(M, N, K):
        A = np.random.rand(m, k).astype(np.float32)
        B = np.random.rand(k, n).astype(np.float32)
        C = A @ B

        C_out = np.zeros_like(C)

        fn(None, m, n, k, C_out, A, B)
        np.testing.assert_almost_equal(C, C_out, decimal=3)


@pytest.fixture
def sgemm_6x16():
    @proc
    def sgemm_6x16(
        K: size,
        C: [f32][6, 16] @ DRAM,
        A: [f32][6, K] @ DRAM,
        B: [f32][K, 16] @ DRAM,
    ):
        for i in seq(0, 6):
            for j in seq(0, 16):
                for k in seq(0, K):
                    C[i, j] += A[i, k] * B[k, j]

    return sgemm_6x16


@pytest.fixture
def avx2_sgemm_6x16(sgemm_6x16):
    avx = rename(sgemm_6x16, "rank_k_reduce_6x16_scheduled")

    avx = stage_mem(avx, "C[_] += _", "C[i, j]", "C_reg")
    avx = set_memory(avx, "C_reg", AVX2)
    avx = old_split(avx, "j", 8, ["jo", "ji"], perfect=True)
    avx = reorder_loops(avx, "ji k")
    avx = reorder_loops(avx, "jo k")
    avx = reorder_loops(avx, "i k")
    avx = autolift_alloc(avx, "C_reg:_", n_lifts=3, keep_dims=True)
    avx = old_fission_after(avx, "C_reg = _ #0", n_lifts=3)
    avx = old_fission_after(avx, "C_reg[_] += _ #0", n_lifts=3)
    avx = autolift_alloc(avx, "C_reg:_", n_lifts=1)
    avx = old_fission_after(avx, "for i in _:_#0", n_lifts=1)
    avx = old_fission_after(avx, "for i in _:_#1", n_lifts=1)
    avx = simplify(avx)

    avx2_sgemm_6x16 = avx

    def sched_avx2_sgemm_6x16(p=avx2_sgemm_6x16):
        p = bind_expr(p, "A[i, k]", "a_vec")
        p = set_memory(p, "a_vec", AVX2)
        p = expand_dim(p, "a_vec:_", "8", "ji")
        p = autolift_alloc(p, "a_vec:_")
        p = old_fission_after(p, "a_vec[_] = _")
        #
        p = bind_expr(p, "B[k, _]", "b_vec")
        p = set_memory(p, "b_vec", AVX2)
        p = expand_dim(p, "b_vec:_", "8", "ji")
        p = autolift_alloc(p, "b_vec:_")
        p = old_fission_after(p, "b_vec[_] = _")
        #
        p = replace_all(p, avx2_set0_ps)
        p = replace_all(p, mm256_broadcast_ss)
        p = replace_all(p, mm256_fmadd_ps)
        p = replace_all(p, avx2_fmadd_memu_ps)
        p = replace(p, "for ji in _:_ #0", mm256_loadu_ps)
        p = replace(p, "for ji in _:_ #0", mm256_loadu_ps)
        p = replace(p, "for ji in _:_ #0", mm256_storeu_ps)
        #
        p = old_unroll(p, "jo")
        p = old_unroll(p, "i")
        #
        p = simplify(p)
        return p

    avx2_sgemm_6x16 = sched_avx2_sgemm_6x16()

    return avx2_sgemm_6x16


# @pytest.mark.skip(reason='apparently unifying the broadcast is '
#                         'non-deterministic')
# Until this non-determinism can be removed, at least
# try to run the code here...
def test_gen_avx2_sgemm_kernel(avx2_sgemm_6x16):
    pass
    # assert str(avx2_sgemm_6x16) == golden


@pytest.fixture
def sgemm_full():
    @proc
    def sgemm_full(
        N: size,
        M: size,
        K: size,
        C: f32[N, M] @ DRAM,
        A: f32[N, K] @ DRAM,
        B: f32[K, M] @ DRAM,
    ):
        assert K > 0

        for i in seq(0, N):
            for j in seq(0, M):
                for k in seq(0, K):
                    C[i, j] += A[i, k] * B[k, j]

    return sgemm_full


@pytest.fixture
def avx2_sgemm_full(sgemm_full, sgemm_6x16, avx2_sgemm_6x16):
    cache_i = 16
    cache_j = 4
    cache_k = 2

    def sched_avx_sgemm_full(p=sgemm_full):
        print(p)
        p = rename(p, "avx_sgemm_full")
        # initial i,j tiling
        p = old_split(p, "i", 6, ["io", "ii"], tail="cut")
        p = reorder_loops(p, "ii j #0")
        p = divide_loop(p, "j #0", 16, ["jo", "ji"], tail="cut")
        p = reorder_loops(p, "ji ii")
        # breaking off the main loop
        p = old_fission_after(p, "for jo in _: _")
        # introduce k-tiling for later
        p = divide_loop(p, "k #0", cache_k * 16, ["ko", "ki"], tail="cut")
        p = old_fission_after(p, "for ko in _: _", n_lifts=2)
        p = reorder_loops(p, "ji ko")
        p = reorder_loops(p, "ii ko")
        p = replace_all(p, sgemm_6x16)
        # insert uses of micro-kernel now
        p = call_eqv(p, "sgemm_6x16(_, _, _, _)", avx2_sgemm_6x16)
        p = call_eqv(p, "sgemm_6x16(_, _, _, _)", avx2_sgemm_6x16)
        # do outer tiling for cache-locality
        p = divide_loop(p, "io #0", cache_i, ["io", "im"], tail="cut")
        p = reorder_loops(p, "im jo")
        p = divide_loop(p, "jo #0", cache_j, ["jo", "jm"], tail="cut")
        p = divide_loop(p, "jo #1", cache_j, ["jo", "jm"], tail="cut")
        p = reorder_loops(p, "jm im")
        # move the ko loop up and out
        p = old_fission_after(p, "for ko in _: _", n_lifts=2)
        p = reorder_loops(p, "jm ko #0")
        p = reorder_loops(p, "im ko #0")
        return p

    avx_sgemm_full = sched_avx_sgemm_full()

    return avx_sgemm_full


# just make sure the scheduling works
def test_gen_avx2_sgemm_full(avx2_sgemm_full):
    pass


@pytest.mark.isa("AVX2")
def test_avx2_sgemm_full(compiler, avx2_sgemm_full):
    fn = compiler.compile(
        avx2_sgemm_full, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    _sgemm_test_cases(
        fn, M=range(10, 600, 200), N=range(20, 400, 120), K=range(1, 512, 160)
    )


@pytest.mark.isa("AVX2")
def test_avx2_sgemm_6x16(compiler, avx2_sgemm_6x16):
    @proc
    def avx2_sgemm_6x16_wrapper(
        M: size,
        N: size,
        K: size,
        C: f32[6, 16] @ DRAM,
        A: f32[6, K] @ DRAM,
        B: f32[K, 16] @ DRAM,
    ):
        avx2_sgemm_6x16(K, C, A, B)

    fn = compiler.compile(
        avx2_sgemm_6x16_wrapper, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    _sgemm_test_cases(fn, M=[6], N=[16], K=range(1, 512))


@pytest.fixture
def spec_kernel(sgemm_full, sgemm_6x16, avx2_sgemm_6x16):
    @proc
    def sgemm_micro_kernel_staged(
        M: size,
        N: size,
        K: size,
        A: f32[M, K],
        B: f32[K, 16 * ((N + 15) / 16)],
        C: [f32][M, N],
    ):
        assert M >= 1
        assert N >= 1
        assert K >= 1
        assert stride(A, 1) == 1
        assert stride(B, 1) == 1
        assert stride(C, 1) == 1

        C_reg: f32[M, ((N + 15) / 16), 16] @ AVX512
        for i in seq(0, M):
            for j in seq(0, N / 16):
                mm512_loadu_ps(C_reg[i, j, :], C[i, 16 * j : 16 * j + 16])
            if N % 16 > 0:
                mm512_maskz_loadu_ps(
                    N % 16,
                    C_reg[i, N / 16, :],
                    C[i, 16 * (N / 16) : 16 * (N / 16) + N % 16],
                )

        for k in seq(0, K):
            for i in seq(0, M):
                a_vec: f32[16] @ AVX512
                mm512_set1_ps(a_vec, A[i, k : k + 1])
                for j in seq(0, ((N + 15) / 16)):
                    b_vec: f32[16] @ AVX512
                    mm512_loadu_ps(b_vec, B[k, j * 16 : j * 16 + 16])
                    mm512_fmadd_ps(a_vec, b_vec, C_reg[i, j, :])

        for i in seq(0, M):
            for j in seq(0, N / 16):
                mm512_storeu_ps(C[i, 16 * j : 16 * j + 16], C_reg[i, j, :])
            if N % 16 > 0:
                mm512_mask_storeu_ps(
                    N % 16,
                    C[i, 16 * (N / 16) : 16 * (N / 16) + N % 16],
                    C_reg[i, N / 16, :],
                )

    def sched_spec_kernel(p=sgemm_micro_kernel_staged):
        p = p.partial_eval(6, 64)
        p = simplify(p)
        p = old_unroll(p, "j")
        p = old_unroll(p, "i")
        p = simplify(p)
        return p

    spec_kernel = sched_spec_kernel()

    return spec_kernel


# just make sure the scheduling works
def test_gen_avx512_sgemm_full(spec_kernel):
    pass


@pytest.mark.isa("AVX512f")
def test_avx512_sgemm_full(compiler, spec_kernel):

    spec_kernel.c_code_str()

    @proc
    def sgemm_full(
        N: size,
        M: size,
        K: size,
        C: f32[N, M] @ DRAM,
        A: f32[N, K] @ DRAM,
        B: f32[K, M] @ DRAM,
    ):
        for i in seq(0, N):
            for j in seq(0, M):
                for k in seq(0, K):
                    C[i, j] += A[i, k] * B[k, j]

    fn = compiler.compile(
        sgemm_full, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake-avx512"
    )

    _sgemm_test_cases(
        fn, M=range(10, 600, 200), N=range(20, 400, 120), K=range(1, 512, 160)
    )


@pytest.fixture
def simple_buffer_select():
    @proc
    def simple_buffer_select(
        N: size,
        out: f32[N] @ DRAM,
        x: f32[N] @ DRAM,
        v: f32[N] @ DRAM,
        y: f32[N] @ DRAM,
        z: f32[N] @ DRAM,
    ):
        assert N >= 1

        for i in seq(0, N):
            out[i] = select(x[i], v[i], y[i], z[i])

    VEC_W = 256 // 32
    outer_it = "io"
    inner_it = "ii"
    loop_fragment = lambda it, idx=0: f"for {it} in _:_ #{idx}"

    def sched_simple_buffer_select(proc):
        proc = divide_loop(
            proc, loop_fragment("i"), VEC_W, (outer_it, inner_it), tail="cut"
        )
        stage = lambda proc, buffer: set_memory(
            stage_mem(
                proc,
                loop_fragment(inner_it),
                f"{buffer}[{VEC_W} * {outer_it}:{VEC_W} * {outer_it} + {VEC_W}]",
                f"{buffer}Reg",
            ),
            f"{buffer}Reg",
            AVX2,
        )
        proc = stage(proc, "x")
        proc = stage(proc, "v")
        proc = stage(proc, "y")
        proc = stage(proc, "z")
        proc = stage(proc, "out")
        proc = replace(proc, loop_fragment("i0"), mm256_loadu_ps)
        proc = replace(proc, loop_fragment("i0"), mm256_loadu_ps)
        proc = replace(proc, loop_fragment("i0"), mm256_loadu_ps)
        proc = replace(proc, loop_fragment("i0"), mm256_loadu_ps)
        proc = replace(proc, loop_fragment(inner_it), avx2_select_ps)
        proc = replace(proc, loop_fragment("i0"), mm256_storeu_ps)
        proc = bind_expr(proc, "x[ii + N / 8 * 8]", "x_temp")
        proc = bind_expr(proc, "v[ii + N / 8 * 8]", "v_temp")
        proc = bind_expr(proc, "y[ii + N / 8 * 8]", "y_temp")
        proc = bind_expr(proc, "z[ii + N / 8 * 8]", "z_temp")
        proc = simplify(proc)

        return proc

    return sched_simple_buffer_select(simple_buffer_select)


@pytest.mark.isa("AVX2")
def test_avx2_select_ps(compiler, simple_buffer_select):
    fn = compiler.compile(
        simple_buffer_select, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    def run_and_check(N, x, v, y, z):
        expected = np.zeros(N, dtype=np.float32)
        for i in range(N):
            if x[i] < v[i]:
                expected[i] = y[i]
            else:
                expected[i] = z[i]

        out = np.array(np.random.rand(N), dtype=np.float32)
        x_copy = x.copy()
        v_copy = v.copy()
        y_copy = y.copy()
        z_copy = z.copy()

        fn(None, N, out, x, v, y, z)

        np.testing.assert_almost_equal(out, expected)
        np.testing.assert_almost_equal(x, x_copy)
        np.testing.assert_almost_equal(v, v_copy)
        np.testing.assert_almost_equal(y, y_copy)
        np.testing.assert_almost_equal(z, z_copy)

    N = 50
    x = np.array([float(i) for i in range(N)], dtype=np.float32)
    v = np.array(
        [1, 0, 3, 4, 1, 1, 9, 0]
        + [100] * 8
        + [0] * 8
        + [float(i) for i in range(24, N)],
        dtype=np.float32,
    )
    y = np.array([float(i) for i in range(200, 200 + N)], dtype=np.float32)
    z = np.array([float(i) for i in range(300, 300 + N)], dtype=np.float32)

    # Correctness testing
    run_and_check(N, x, v, y, z)

    # Precision testing
    N = 111
    run_and_check(
        N,
        np.array(np.random.rand(N), dtype=np.float32),
        np.array(np.random.rand(N), dtype=np.float32),
        np.array(np.random.rand(N), dtype=np.float32),
        np.array(np.random.rand(N), dtype=np.float32),
    )


@pytest.mark.isa("AVX2")
def test_avx2_assoc_reduce_add_ps(compiler):
    @proc
    def accumulate_buffer(x: f32[8], result: [f32][1]):
        tmp_result: f32
        tmp_result = result[0]
        for i in seq(0, 8):
            tmp_result += x[i]
        result[0] = tmp_result

    accumulate_buffer = stage_mem(accumulate_buffer, "for i in _:_", "x[0:8]", "xReg")
    accumulate_buffer = set_memory(accumulate_buffer, "xReg", AVX2)
    accumulate_buffer = replace(accumulate_buffer, "for i0 in _:_", mm256_loadu_ps)
    accumulate_buffer = replace(
        accumulate_buffer, "for i in _:_", avx2_assoc_reduce_add_ps
    )
    accumulate_buffer = simplify(accumulate_buffer)

    fn = compiler.compile(
        accumulate_buffer, skip_on_fail=True, CMAKE_C_FLAGS="-march=skylake"
    )

    def run_and_check(x, result):
        expected = result.copy()
        for i in range(8):
            expected[0] += x[i]

        x_copy = x.copy()

        fn(None, x, result)

        # lower precision checking because we are assuming float addition is associative in the instruction
        np.testing.assert_almost_equal(result, expected, decimal=4)
        np.testing.assert_almost_equal(x, x_copy)

    result = np.array([0.0], dtype=np.float32)
    x = np.array([1, 2, 3, 1, 2, 3, 7, 2], dtype=np.float32)

    fn(None, x, result)

    run_and_check(x, result)
    run_and_check(
        np.array(
            [
                0.47299325,
                0.2869141,
                0.23663807,
                0.12012372,
                0.93651915,
                0.06829825,
                0.22391547,
                0.20829211,
            ],
            dtype=np.float32,
        ),
        np.array([0.11827405109974021], dtype=np.float32),
    )


@pytest.mark.isa("AVX2")
def test_mm256_broadcast_ss_scalar(compiler):
    @proc
    def mm256_broadcast_ss_scalar_wrapper(out: f32[8] @ DRAM, val: f32[1] @ DRAM):
        tmp_val: f32
        tmp_val = val[0]
        tmp_buffer_0: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_0, out)
        mm256_broadcast_ss_scalar(tmp_buffer_0, tmp_val)
        mm256_storeu_ps(out, tmp_buffer_0)
        val[0] = tmp_val

    @proc
    def mm256_broadcast_ss_scalar_ref(out: f32[8] @ DRAM, val: f32[1] @ DRAM):
        assert stride(out, 0) == 1
        for i in seq(0, 8):
            out[i] = val[0]

    fn = compiler.compile(
        [mm256_broadcast_ss_scalar_wrapper, mm256_broadcast_ss_scalar_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )

    out = np.array(
        [
            0.8352003,
            0.119042955,
            0.05500923,
            0.9115426,
            0.8030574,
            0.07465152,
            0.5667018,
            0.53294945,
        ],
        dtype=np.float32,
    )
    val = np.array([2.44])
    out_copy = out.copy()
    val_copy = val.copy()

    getattr(fn, "mm256_broadcast_ss_scalar_wrapper")(None, out, val)
    getattr(fn, "mm256_broadcast_ss_scalar_ref")(None, out_copy, val_copy)

    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(val, val_copy)


@pytest.mark.isa("AVX2")
def test_mm256_add_ps(compiler):
    @proc
    def mm256_add_ps_wrapper(out: f32[8] @ DRAM, x: f32[8] @ DRAM, y: f32[8] @ DRAM):
        tmp_buffer_0: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_0, out)
        tmp_buffer_1: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_1, x)
        tmp_buffer_2: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_2, y)
        mm256_add_ps(tmp_buffer_0, tmp_buffer_1, tmp_buffer_2)
        mm256_storeu_ps(out, tmp_buffer_0)
        mm256_storeu_ps(x, tmp_buffer_1)
        mm256_storeu_ps(y, tmp_buffer_2)

    @proc
    def mm256_add_ps_ref(out: f32[8] @ DRAM, x: f32[8] @ DRAM, y: f32[8] @ DRAM):
        assert stride(out, 0) == 1
        assert stride(x, 0) == 1
        assert stride(y, 0) == 1
        for i in seq(0, 8):
            out[i] = x[i] + y[i]

    fn = compiler.compile(
        [mm256_add_ps_wrapper, mm256_add_ps_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )

    out = np.array(
        [
            0.19168626,
            0.9292728,
            0.08044847,
            0.118411385,
            0.586027,
            0.9493457,
            0.91186064,
            0.97682995,
        ],
        dtype=np.float32,
    )
    x = np.array(
        [
            0.29922104,
            0.6485327,
            0.111039855,
            0.12260633,
            0.86832726,
            0.06105361,
            0.26866043,
            0.38205943,
        ],
        dtype=np.float32,
    )
    y = np.array(
        [
            0.30989048,
            0.3230521,
            0.70411354,
            0.117873766,
            0.5853253,
            0.08196206,
            0.31477037,
            0.8284393,
        ],
        dtype=np.float32,
    )
    out_copy = out.copy()
    x_copy = x.copy()
    y_copy = y.copy()

    getattr(fn, "mm256_add_ps_wrapper")(None, out, x, y)
    getattr(fn, "mm256_add_ps_ref")(None, out_copy, x_copy, y_copy)

    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(x, x_copy)
    np.testing.assert_almost_equal(y, y_copy)


@pytest.mark.isa("AVX2")
def test_avx2_reg_copy_ps(compiler):
    @proc
    def avx2_reg_copy_ps_wrapper(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        tmp_buffer_0: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_0, dst)
        tmp_buffer_1: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_1, src)
        avx2_reg_copy_ps(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_ps(dst, tmp_buffer_0)
        mm256_storeu_ps(src, tmp_buffer_1)

    @proc
    def avx2_reg_copy_ps_ref(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 8):
            dst[i] = src[i]

    fn = compiler.compile(
        [avx2_reg_copy_ps_wrapper, avx2_reg_copy_ps_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )

    dst = np.array(
        [
            0.95774984,
            0.5476575,
            0.2301955,
            0.69115007,
            0.6172301,
            0.37313432,
            0.5556015,
            0.7339064,
        ],
        dtype=np.float32,
    )
    src = np.array(
        [
            0.29309168,
            0.034619242,
            0.0769644,
            0.9533431,
            0.11209598,
            0.5699761,
            0.36666384,
            0.32560244,
        ],
        dtype=np.float32,
    )
    dst_copy = dst.copy()
    src_copy = src.copy()

    getattr(fn, "avx2_reg_copy_ps_wrapper")(None, dst, src)
    getattr(fn, "avx2_reg_copy_ps_ref")(None, dst_copy, src_copy)

    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


@pytest.mark.isa("AVX2")
def test_avx2_sign_ps(compiler):
    @proc
    def avx2_sign_ps_wrapper(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        tmp_buffer_0: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_0, dst)
        tmp_buffer_1: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_1, src)
        avx2_sign_ps(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_ps(dst, tmp_buffer_0)
        mm256_storeu_ps(src, tmp_buffer_1)

    @proc
    def avx2_sign_ps_ref(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 8):
            dst[i] = -src[i]

    fn = compiler.compile(
        [avx2_sign_ps_wrapper, avx2_sign_ps_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )

    dst = np.array(
        [
            0.61119187,
            0.57439226,
            0.63750356,
            0.01567109,
            0.7531479,
            0.80388564,
            0.6817162,
            0.3611551,
        ],
        dtype=np.float32,
    )
    src = np.array(
        [
            0.1869111,
            0.53224814,
            0.71396947,
            0.7539144,
            0.6865989,
            0.33050302,
            0.86975175,
            0.17079325,
        ],
        dtype=np.float32,
    )

    dst_copy = dst.copy()
    src_copy = src.copy()

    getattr(fn, "avx2_sign_ps_wrapper")(None, dst, src)
    getattr(fn, "avx2_sign_ps_ref")(None, dst_copy, src_copy)

    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


@pytest.mark.isa("AVX2")
def test_avx2_reduce_add_wide_ps(compiler):
    @proc
    def avx2_reduce_add_wide_ps_wrapper(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        tmp_buffer_0: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_0, dst)
        tmp_buffer_1: f32[8] @ AVX2
        mm256_loadu_ps(tmp_buffer_1, src)
        avx2_reduce_add_wide_ps(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_ps(dst, tmp_buffer_0)
        mm256_storeu_ps(src, tmp_buffer_1)

    @proc
    def avx2_reduce_add_wide_ps_ref(dst: f32[8] @ DRAM, src: f32[8] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 8):
            dst[i] += src[i]

    fn = compiler.compile(
        [avx2_reduce_add_wide_ps_wrapper, avx2_reduce_add_wide_ps_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )

    dst = np.array(
        [
            0.82633126,
            0.18224466,
            0.8483131,
            0.85528636,
            0.9373481,
            0.87415653,
            0.619115,
            0.85448426,
        ],
        dtype=np.float32,
    )
    src = np.array(
        [
            0.3565467,
            0.25555333,
            0.8524338,
            0.33920884,
            0.5461596,
            0.93643206,
            0.7152863,
            0.7914703,
        ],
        dtype=np.float32,
    )
    dst_copy = dst.copy()
    src_copy = src.copy()

    getattr(fn, "avx2_reduce_add_wide_ps_wrapper")(None, dst, src)
    getattr(fn, "avx2_reduce_add_wide_ps_ref")(None, dst_copy, src_copy)

    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


@pytest.mark.isa("AVX2")
def test_mm256_setzero_pd(compiler):
    @proc
    def mm256_setzero_pd_wrapper(dst: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, dst)
        mm256_setzero_pd(tmp_buffer_0)
        mm256_storeu_pd(dst, tmp_buffer_0)

    @proc
    def mm256_setzero_pd_ref(dst: f64[4] @ DRAM):
        assert stride(dst, 0) == 1
        for i in seq(0, 4):
            dst[i] = 0.0

    fn = compiler.compile(
        [mm256_setzero_pd_wrapper, mm256_setzero_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    dst = np.array(
        [
            0.7702997103754631,
            0.45626842944141055,
            0.7439364522403988,
            0.701510512624819,
        ],
        dtype=np.float64,
    )
    dst_copy = dst.copy()
    getattr(fn, "mm256_setzero_pd_wrapper")(None, dst)
    getattr(fn, "mm256_setzero_pd_ref")(None, dst_copy)
    np.testing.assert_almost_equal(dst, dst_copy)


@pytest.mark.isa("AVX2")
def test_mm256_fmadd_pd(compiler):
    @proc
    def mm256_fmadd_pd_wrapper(
        dst: f64[4] @ DRAM, src1: f64[4] @ DRAM, src2: f64[4] @ DRAM
    ):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, dst)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, src1)
        tmp_buffer_2: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_2, src2)
        mm256_fmadd_pd(tmp_buffer_0, tmp_buffer_1, tmp_buffer_2)
        mm256_storeu_pd(dst, tmp_buffer_0)
        mm256_storeu_pd(src1, tmp_buffer_1)
        mm256_storeu_pd(src2, tmp_buffer_2)

    @proc
    def mm256_fmadd_pd_ref(
        dst: f64[4] @ DRAM, src1: f64[4] @ DRAM, src2: f64[4] @ DRAM
    ):
        assert stride(src1, 0) == 1
        assert stride(src2, 0) == 1
        assert stride(dst, 0) == 1
        for i in seq(0, 4):
            dst[i] += src1[i] * src2[i]

    fn = compiler.compile(
        [mm256_fmadd_pd_wrapper, mm256_fmadd_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    dst = np.array(
        [
            0.06690474560928039,
            0.0850078709985519,
            0.029640255722781395,
            0.7284983930187073,
        ],
        dtype=np.float64,
    )
    src1 = np.array(
        [
            0.16333810334859344,
            0.773775867984533,
            0.4243993631288452,
            0.03447281200342622,
        ],
        dtype=np.float64,
    )
    src2 = np.array(
        [
            0.6898950978728396,
            0.7565890349219402,
            0.9694891176666618,
            0.18128624734877385,
        ],
        dtype=np.float64,
    )
    dst_copy = dst.copy()
    src1_copy = src1.copy()
    src2_copy = src2.copy()
    getattr(fn, "mm256_fmadd_pd_wrapper")(None, dst, src1, src2)
    getattr(fn, "mm256_fmadd_pd_ref")(None, dst_copy, src1_copy, src2_copy)
    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src1, src1_copy)
    np.testing.assert_almost_equal(src2, src2_copy)


@pytest.mark.isa("AVX2")
def test_mm256_broadcast_sd(compiler):
    @proc
    def mm256_broadcast_sd_wrapper(out: f64[4] @ DRAM, val: f64[1] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, out)
        mm256_broadcast_sd(tmp_buffer_0, val)
        mm256_storeu_pd(out, tmp_buffer_0)

    @proc
    def mm256_broadcast_sd_ref(out: f64[4] @ DRAM, val: f64[1] @ DRAM):
        assert stride(out, 0) == 1
        for i in seq(0, 4):
            out[i] = val[0]

    fn = compiler.compile(
        [mm256_broadcast_sd_wrapper, mm256_broadcast_sd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    out = np.array(
        [
            0.9992828001754692,
            0.6054592176163047,
            0.21423730227883409,
            0.06414175723973703,
        ],
        dtype=np.float64,
    )
    val = np.array([0.6716955229801979], dtype=np.float64)
    out_copy = out.copy()
    val_copy = val.copy()
    getattr(fn, "mm256_broadcast_sd_wrapper")(None, out, val)
    getattr(fn, "mm256_broadcast_sd_ref")(None, out_copy, val_copy)
    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(val, val_copy)


@pytest.mark.isa("AVX2")
def test_mm256_broadcast_sd_scalar(compiler):
    @proc
    def mm256_broadcast_sd_scalar_wrapper(out: f64[4] @ DRAM, val: f64[1] @ DRAM):
        tmp_val: f64
        tmp_val = val[0]
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, out)
        mm256_broadcast_sd_scalar(tmp_buffer_0, tmp_val)
        mm256_storeu_pd(out, tmp_buffer_0)

    @proc
    def mm256_broadcast_sd_scalar_ref(out: f64[4] @ DRAM, val: f64[1] @ DRAM):
        assert stride(out, 0) == 1
        for i in seq(0, 4):
            out[i] = val[0]

    fn = compiler.compile(
        [mm256_broadcast_sd_scalar_wrapper, mm256_broadcast_sd_scalar_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    out = np.array(
        [
            0.08715297607238981,
            0.2554485029748307,
            0.8467682976066226,
            0.9005310811939783,
        ],
        dtype=np.float64,
    )
    val = np.array([3.51])
    out_copy = out.copy()
    val_copy = val.copy()
    getattr(fn, "mm256_broadcast_sd_scalar_wrapper")(None, out, val)
    getattr(fn, "mm256_broadcast_sd_scalar_ref")(None, out_copy, val_copy)
    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(val, val_copy)


@pytest.mark.isa("AVX2")
def test_mm256_mul_pd(compiler):
    @proc
    def mm256_mul_pd_wrapper(out: f64[4] @ DRAM, x: f64[4] @ DRAM, y: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, out)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, x)
        tmp_buffer_2: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_2, y)
        mm256_mul_pd(tmp_buffer_0, tmp_buffer_1, tmp_buffer_2)
        mm256_storeu_pd(out, tmp_buffer_0)
        mm256_storeu_pd(x, tmp_buffer_1)
        mm256_storeu_pd(y, tmp_buffer_2)

    @proc
    def mm256_mul_pd_ref(out: f64[4] @ DRAM, x: f64[4] @ DRAM, y: f64[4] @ DRAM):
        assert stride(out, 0) == 1
        assert stride(x, 0) == 1
        assert stride(y, 0) == 1
        for i in seq(0, 4):
            out[i] = x[i] * y[i]

    fn = compiler.compile(
        [mm256_mul_pd_wrapper, mm256_mul_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    out = np.array(
        [
            0.20493105389880395,
            0.9358746107512442,
            0.5513072892432449,
            0.1416622828015115,
        ],
        dtype=np.float64,
    )
    x = np.array(
        [
            0.6359108492392163,
            0.7391629379581788,
            0.8894695567639811,
            0.6009249216013818,
        ],
        dtype=np.float64,
    )
    y = np.array(
        [
            0.9102732096226585,
            0.6246662381685367,
            0.8297542680325571,
            0.7470065913040405,
        ],
        dtype=np.float64,
    )
    out_copy = out.copy()
    x_copy = x.copy()
    y_copy = y.copy()
    getattr(fn, "mm256_mul_pd_wrapper")(None, out, x, y)
    getattr(fn, "mm256_mul_pd_ref")(None, out_copy, x_copy, y_copy)
    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(x, x_copy)
    np.testing.assert_almost_equal(y, y_copy)


@pytest.mark.isa("AVX2")
def test_mm256_add_pd(compiler):
    @proc
    def mm256_add_pd_wrapper(out: f64[4] @ DRAM, x: f64[4] @ DRAM, y: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, out)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, x)
        tmp_buffer_2: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_2, y)
        mm256_add_pd(tmp_buffer_0, tmp_buffer_1, tmp_buffer_2)
        mm256_storeu_pd(out, tmp_buffer_0)
        mm256_storeu_pd(x, tmp_buffer_1)
        mm256_storeu_pd(y, tmp_buffer_2)

    @proc
    def mm256_add_pd_ref(out: f64[4] @ DRAM, x: f64[4] @ DRAM, y: f64[4] @ DRAM):
        assert stride(out, 0) == 1
        assert stride(x, 0) == 1
        assert stride(y, 0) == 1
        for i in seq(0, 4):
            out[i] = x[i] + y[i]

    fn = compiler.compile(
        [mm256_add_pd_wrapper, mm256_add_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    out = np.array(
        [
            0.8527647836251072,
            0.4656462939895032,
            0.35186540014581635,
            0.630313690114371,
        ],
        dtype=np.float64,
    )
    x = np.array(
        [
            0.24068467638706337,
            0.8464360008490506,
            0.783023994276535,
            0.7109937138465194,
        ],
        dtype=np.float64,
    )
    y = np.array(
        [
            0.9352666294024309,
            0.32475739858813224,
            0.15796304969982733,
            0.31789446356547857,
        ],
        dtype=np.float64,
    )
    out_copy = out.copy()
    x_copy = x.copy()
    y_copy = y.copy()
    getattr(fn, "mm256_add_pd_wrapper")(None, out, x, y)
    getattr(fn, "mm256_add_pd_ref")(None, out_copy, x_copy, y_copy)
    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(x, x_copy)
    np.testing.assert_almost_equal(y, y_copy)


@pytest.mark.isa("AVX2")
def test_avx2_select_pd(compiler):
    @proc
    def avx2_select_pd_wrapper(
        out: f64[4] @ DRAM,
        x: f64[4] @ DRAM,
        v: f64[4] @ DRAM,
        y: f64[4] @ DRAM,
        z: f64[4] @ DRAM,
    ):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, out)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, x)
        tmp_buffer_2: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_2, v)
        tmp_buffer_3: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_3, y)
        tmp_buffer_4: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_4, z)
        avx2_select_pd(
            tmp_buffer_0, tmp_buffer_1, tmp_buffer_2, tmp_buffer_3, tmp_buffer_4
        )
        mm256_storeu_pd(out, tmp_buffer_0)
        mm256_storeu_pd(x, tmp_buffer_1)
        mm256_storeu_pd(v, tmp_buffer_2)
        mm256_storeu_pd(y, tmp_buffer_3)
        mm256_storeu_pd(z, tmp_buffer_4)

    @proc
    def avx2_select_pd_ref(
        out: f64[4] @ DRAM,
        x: f64[4] @ DRAM,
        v: f64[4] @ DRAM,
        y: f64[4] @ DRAM,
        z: f64[4] @ DRAM,
    ):
        assert stride(out, 0) == 1
        assert stride(x, 0) == 1
        assert stride(v, 0) == 1
        assert stride(y, 0) == 1
        assert stride(z, 0) == 1
        for i in seq(0, 4):
            xTmp: f64
            xTmp = x[i]
            vTmp: f64
            vTmp = v[i]
            yTmp: f64
            yTmp = y[i]
            zTmp: f64
            zTmp = z[i]
            out[i] = select(xTmp, vTmp, yTmp, zTmp)

    fn = compiler.compile(
        [avx2_select_pd_wrapper, avx2_select_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    out = np.array(
        [
            0.15072544011156797,
            0.36126593942879226,
            0.2962307720690991,
            0.5533859089478346,
        ],
        dtype=np.float64,
    )
    x = np.array(
        [
            0.8735799415726602,
            0.29340630006863155,
            0.3845715188628306,
            0.04794077205280567,
        ],
        dtype=np.float64,
    )
    v = np.array(
        [
            0.47764878514866593,
            0.7121357194784308,
            0.08003544890580039,
            0.16599111414940448,
        ],
        dtype=np.float64,
    )
    y = np.array(
        [
            0.18912116677473167,
            0.012455292110266747,
            0.9069633225674659,
            0.5928029432231223,
        ],
        dtype=np.float64,
    )
    z = np.array(
        [
            0.4357337303927791,
            0.7667888372979041,
            0.6989032832260329,
            0.006324755999849274,
        ],
        dtype=np.float64,
    )
    out_copy = out.copy()
    x_copy = x.copy()
    v_copy = v.copy()
    y_copy = y.copy()
    z_copy = z.copy()
    getattr(fn, "avx2_select_pd_wrapper")(None, out, x, v, y, z)
    getattr(fn, "avx2_select_pd_ref")(None, out_copy, x_copy, v_copy, y_copy, z_copy)
    np.testing.assert_almost_equal(out, out_copy)
    np.testing.assert_almost_equal(x, x_copy)
    np.testing.assert_almost_equal(v, v_copy)
    np.testing.assert_almost_equal(y, y_copy)
    np.testing.assert_almost_equal(z, z_copy)


@pytest.mark.isa("AVX2")
def test_avx2_assoc_reduce_add_pd(compiler):
    @proc
    def avx2_assoc_reduce_add_pd_wrapper(x: f64[4] @ DRAM, result: f64[1] @ DRAM):
        tmp_result: f64
        tmp_result = result[0]
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, x)
        avx2_assoc_reduce_add_pd(tmp_buffer_0, tmp_result)
        mm256_storeu_pd(x, tmp_buffer_0)
        result[0] = tmp_result

    @proc
    def avx2_assoc_reduce_add_pd_ref(x: f64[4] @ DRAM, result: f64[1] @ DRAM):
        assert stride(x, 0) == 1
        for i in seq(0, 4):
            result[0] += x[i]

    fn = compiler.compile(
        [avx2_assoc_reduce_add_pd_wrapper, avx2_assoc_reduce_add_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    x = np.array(
        [
            0.24349706300899643,
            0.44304522832871585,
            0.7845959927912908,
            0.27109035501352363,
        ],
        dtype=np.float64,
    )
    result = np.array([-1.253])
    x_copy = x.copy()
    result_copy = result.copy()
    getattr(fn, "avx2_assoc_reduce_add_pd_wrapper")(None, x, result)
    getattr(fn, "avx2_assoc_reduce_add_pd_ref")(None, x_copy, result_copy)
    np.testing.assert_almost_equal(x, x_copy)
    np.testing.assert_almost_equal(result, result_copy)


@pytest.mark.isa("AVX2")
def test_avx2_sign_pd(compiler):
    @proc
    def avx2_sign_pd_wrapper(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, dst)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, src)
        avx2_sign_pd(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_pd(dst, tmp_buffer_0)
        mm256_storeu_pd(src, tmp_buffer_1)

    @proc
    def avx2_sign_pd_ref(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 4):
            dst[i] = -src[i]

    fn = compiler.compile(
        [avx2_sign_pd_wrapper, avx2_sign_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    dst = np.array(
        [
            0.066282086666972,
            0.23414881487540895,
            0.4922961364730185,
            0.11810813857973845,
        ],
        dtype=np.float64,
    )
    src = np.array(
        [
            0.06436446994101086,
            0.637838437168881,
            0.6797374431407577,
            0.040744738211098475,
        ],
        dtype=np.float64,
    )
    dst_copy = dst.copy()
    src_copy = src.copy()
    getattr(fn, "avx2_sign_pd_wrapper")(None, dst, src)
    getattr(fn, "avx2_sign_pd_ref")(None, dst_copy, src_copy)
    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


@pytest.mark.isa("AVX2")
def test_avx2_reduce_add_wide_pd(compiler):
    @proc
    def avx2_reduce_add_wide_pd_wrapper(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, dst)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, src)
        avx2_reduce_add_wide_pd(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_pd(dst, tmp_buffer_0)
        mm256_storeu_pd(src, tmp_buffer_1)

    @proc
    def avx2_reduce_add_wide_pd_ref(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 4):
            dst[i] += src[i]

    fn = compiler.compile(
        [avx2_reduce_add_wide_pd_wrapper, avx2_reduce_add_wide_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    dst = np.array(
        [
            0.15640257715569084,
            0.6783560522800346,
            0.3288580779472433,
            0.05681854279921417,
        ],
        dtype=np.float64,
    )
    src = np.array(
        [
            0.9117144753906365,
            0.6016906956921138,
            0.3095551646830016,
            0.5330758995820801,
        ],
        dtype=np.float64,
    )
    dst_copy = dst.copy()
    src_copy = src.copy()
    getattr(fn, "avx2_reduce_add_wide_pd_wrapper")(None, dst, src)
    getattr(fn, "avx2_reduce_add_wide_pd_ref")(None, dst_copy, src_copy)
    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


@pytest.mark.isa("AVX2")
def test_avx2_reg_copy_pd(compiler):
    @proc
    def avx2_reg_copy_pd_wrapper(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        tmp_buffer_0: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_0, dst)
        tmp_buffer_1: f64[4] @ AVX2
        mm256_loadu_pd(tmp_buffer_1, src)
        avx2_reg_copy_pd(tmp_buffer_0, tmp_buffer_1)
        mm256_storeu_pd(dst, tmp_buffer_0)
        mm256_storeu_pd(src, tmp_buffer_1)

    @proc
    def avx2_reg_copy_pd_ref(dst: f64[4] @ DRAM, src: f64[4] @ DRAM):
        assert stride(dst, 0) == 1
        assert stride(src, 0) == 1
        for i in seq(0, 4):
            dst[i] = src[i]

    fn = compiler.compile(
        [avx2_reg_copy_pd_wrapper, avx2_reg_copy_pd_ref],
        skip_on_fail=True,
        CMAKE_C_FLAGS="-march=skylake",
    )
    dst = np.array(
        [
            0.8230485448185485,
            0.400123708278312,
            0.49373866892708396,
            0.7150382338404423,
        ],
        dtype=np.float64,
    )
    src = np.array(
        [
            0.15711822012981025,
            0.9644952348521756,
            0.770817397283775,
            0.9775115541309777,
        ],
        dtype=np.float64,
    )
    dst_copy = dst.copy()
    src_copy = src.copy()
    getattr(fn, "avx2_reg_copy_pd_wrapper")(None, dst, src)
    getattr(fn, "avx2_reg_copy_pd_ref")(None, dst_copy, src_copy)
    np.testing.assert_almost_equal(dst, dst_copy)
    np.testing.assert_almost_equal(src, src_copy)


def test_avx2_divide_by_3(golden):
    @proc
    def foo():
        out: ui16[16] @ AVX2
        x: ui16[16] @ AVX2

        for i in seq(0, 16):
            out[i] = x[i] / 3.0

    foo = replace_all(foo, [avx2_ui16_divide_by_3])
    assert str(foo) == golden



import math


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# BuiltIn superclass


class BuiltIn_Typecheck_Error(Exception):
    def __init__(self, msg):
        self._builtin_err_msg = str(msg)

    def __str__(self):
        return self._builtin_err_msg


_BErr = BuiltIn_Typecheck_Error


class BuiltIn:
    def __init__(self, name):
        self._name = name

    def name(self):
        return self._name

    def globl(self):
        raise NotImplementedError()

    def typecheck(self, args):
        raise NotImplementedError()

    def interpret(self, args):
        raise NotImplementedError()

    def compile(self, args):
        raise NotImplementedError()


class _Sin(BuiltIn):
    def __init__(self):
        super().__init__("sin")

    def typecheck(self, args):
        if len(args) != 1:
            raise _BErr(f"expected 1 argument, got {len(args)}")

        atyp = args[0].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 1 to be a real scalar value, but "
                f"got type {atyp}"
            )
        return atyp

    def globl(self):
        return "#include <math.h>"

    def interpret(self, args):
        return math.sin(args[0])

    def compile(self, args):
        return f"sin((double)*{args[0]})"


sin = _Sin()


class _Relu(BuiltIn):
    def __init__(self):
        super().__init__("relu")

    def typecheck(self, args):
        if len(args) != 1:
            raise _BErr(f"expected 1 argument, got {len(args)}")

        atyp = args[0].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 1 to be a real scalar value, but "
                f"got type {atyp}"
            )
        return atyp

    def globl(self):
        s = (
            "double _relu_(double x) {\n"
            "    if (x > 0.0) return x;\n"
            "    else return 0.0;\n"
            "}\n"
        )
        return s

    def interpret(self, args):
        if args[0] > 0:
            return args[0]
        else:
            return 0

    def compile(self, args):
        return f"_relu_((double)*{args[0]})"


relu = _Relu()


class _Select(BuiltIn):
    def __init__(self):
        super().__init__("select")

    def typecheck(self, args):
        if len(args) != 4:
            raise _BErr(f"expected 4 arguments, got {len(args)}")

        atyp = args[0].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 1 to be a real scalar value, but "
                f"got type {atyp}"
            )

        atyp = args[1].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 2 to be a real scalar value, but "
                f"got type {atyp}"
            )

        atyp = args[2].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 3 to be a real scalar value, but "
                f"got type {atyp}"
            )

        atyp = args[3].type
        if not atyp.is_real_scalar():
            raise _BErr(
                f"expected argument 4 to be a real scalar value, but "
                f"got type {atyp}"
            )
        return atyp

    def globl(self):
        s = (
            "double _select_(double x, double v, double y, double z) {\n"
            "    if (x < v) return y;\n"
            "    else return z;\n"
            "}\n"
        )
        return s

    def interpret(self, args):
        x = args[0]
        v = args[1]
        y = args[2]
        z = args[3]
        if x < v:
            return y
        else:
            return z

    def compile(self, args):
        return f"_select_((double)*{args[0]}, (double)*{args[1]}, (double)*{args[2]}, (double)*{args[3]})"


select = _Select()
